from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.model_selection import train_test_split
from tensorflow import feature_column
from tensorflow.python.client import session
from tensorflow.python.feature_column import dense_features
from tensorflow.python.feature_column import feature_column_v2 as fc
from tensorflow.python.feature_column import sequence_feature_column as sfc
from tensorflow.python.feature_column import serialization
from tensorflow.python.framework import dtypes
from tensorflow.python.framework import ops, errors
from tensorflow.python.framework import sparse_tensor
from tensorflow.python.keras.engine.base_layer import Layer
from tensorflow.python.ops import lookup_ops
from tensorflow.python.ops import math_ops, sparse_ops
from tensorflow.python.ops import variables as variables_lib
from tensorflow.python.platform import test

test_case = test.TestCase()


def _assert_sparse_tensor_indices_shape(expected, actual):
    test_case.assertEqual(np.int64, np.array(actual.indices).dtype)
    test_case.assertAllEqual(expected.indices, actual.indices)

    test_case.assertEqual(np.int64, np.array(actual.dense_shape).dtype)
    test_case.assertAllEqual(expected.dense_shape, actual.dense_shape)


def _assert_sparse_tensor_value(expected, actual):
    _assert_sparse_tensor_indices_shape(expected, actual)

    test_case.assertEqual(np.array(expected.values).dtype, np.array(actual.values).dtype)
    test_case.assertAllEqual(expected.values, actual.values)


def _get_sequence_dense_tensor(column, features):
    return column.get_sequence_dense_tensor(fc.FeatureTransformationCache(features), None)


def _get_sequence_dense_tensor_state(column, features):
    state_manager = fc._StateManagerImpl(Layer(), trainable=True)
    column.create_state(state_manager)
    dense_tensor, lengths = column.get_sequence_dense_tensor(fc.FeatureTransformationCache(features), state_manager)
    return dense_tensor, lengths, state_manager


def _get_sparse_tensors(column, features):
    return column.get_sparse_tensors(
        fc.FeatureTransformationCache(features), None)


def _initialized_session(config=None):
    sess = session.Session(config=config)
    sess.run(variables_lib.global_variables_initializer())
    sess.run(lookup_ops.tables_initializer())
    return sess


class SequenceFeaturesTes:
    def run(self):
        params = [
            {'testcase_name': '2D',
             'sparse_input_args_a': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (2, 0, 1),
                 'dense_shape': (2, 2)},
             'sparse_input_args_b': {
                 # example 0, ids [1]
                 # example 1, ids [2, 0]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (1, 2, 0),
                 'dense_shape': (2, 2)},
             'expected_input_layer': [
                 # example 0, ids_a [2], ids_b [1]
                 [[5., 6., 14., 15., 16.], [0., 0., 0., 0., 0.]],
                 # example 1, ids_a [0, 1], ids_b [2, 0]
                 [[1., 2., 17., 18., 19.], [3., 4., 11., 12., 13.]], ],
             'expected_sequence_length': [1, 2]},
            {'testcase_name': '3D',
             'sparse_input_args_a': {
                 # feature 0, ids [[2], [0, 1]]
                 # feature 1, ids [[0, 0], [1]]
                 'indices': (
                     (0, 0, 0), (0, 1, 0), (0, 1, 1),
                     (1, 0, 0), (1, 0, 1), (1, 1, 0)),
                 'values': (2, 0, 1, 0, 0, 1),
                 'dense_shape': (2, 2, 2)},
             'sparse_input_args_b': {
                 # feature 0, ids [[1, 1], [1]]
                 # feature 1, ids [[2], [0]]
                 'indices': ((0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 1, 0)),
                 'values': (1, 1, 1, 2, 0),
                 'dense_shape': (2, 2, 2)},
             'expected_input_layer': [
                 # feature 0, [a: 2, -, b: 1, 1], [a: 0, 1, b: 1, -]
                 [[5., 6., 14., 15., 16.], [2., 3., 14., 15., 16.]],
                 # feature 1, [a: 0, 0, b: 2, -], [a: 1, -, b: 0, -]
                 [[1., 2., 17., 18., 19.], [3., 4., 11., 12., 13.]]],
             'expected_sequence_length': [2, 2]},
        ]
        for param in params:
            self._embedding_column(
                sparse_input_args_a=param['sparse_input_args_a'],
                sparse_input_args_b=param['sparse_input_args_b'],
                expected_input_layer=param['expected_input_layer'],
                expected_sequence_length=param['expected_sequence_length'],
            )

        params = [
            {'testcase_name': '2D',
             'sparse_input_args_a': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (2, 0, 1),
                 'dense_shape': (2, 2)},
             'sparse_input_args_b': {
                 # example 0, ids [1]
                 # example 1, ids [1, 0]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (1, 1, 0),
                 'dense_shape': (2, 2)},
             'expected_input_layer': [
                 # example 0, ids_a [2], ids_b [1]
                 [[0., 0., 1., 0., 1.], [0., 0., 0., 0., 0.]],
                 # example 1, ids_a [0, 1], ids_b [1, 0]
                 [[1., 0., 0., 0., 1.], [0., 1., 0., 1., 0.]]],
             'expected_sequence_length': [1, 2]},
            {'testcase_name': '3D',
             'sparse_input_args_a': {
                 # feature 0, ids [[2], [0, 1]]
                 # feature 1, ids [[0, 0], [1]]
                 'indices': (
                     (0, 0, 0), (0, 1, 0), (0, 1, 1),
                     (1, 0, 0), (1, 0, 1), (1, 1, 0)),
                 'values': (2, 0, 1, 0, 0, 1),
                 'dense_shape': (2, 2, 2)},
             'sparse_input_args_b': {
                 # feature 0, ids [[1, 1], [1]]
                 # feature 1, ids [[1], [0]]
                 'indices': ((0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 1, 0)),
                 'values': (1, 1, 1, 1, 0),
                 'dense_shape': (2, 2, 2)},
             'expected_input_layer': [
                 # feature 0, [a: 2, -, b: 1, 1], [a: 0, 1, b: 1, -]
                 [[0., 0., 1., 0., 2.], [1., 1., 0., 0., 1.]],
                 # feature 1, [a: 0, 0, b: 1, -], [a: 1, -, b: 0, -]
                 [[2., 0., 0., 0., 1.], [0., 1., 0., 1., 0.]]],
             'expected_sequence_length': [2, 2]},
        ]
        for param in params:
            self._indicator_column(
                sparse_input_args_a=param['sparse_input_args_a'],
                sparse_input_args_b=param['sparse_input_args_b'],
                expected_input_layer=param['expected_input_layer'],
                expected_sequence_length=param['expected_sequence_length'],
            )

        params = [
            {'testcase_name': '2D',
             'sparse_input_args': {
                 # example 0, values [0., 1]
                 # example 1, [10.]
                 'indices': ((0, 0), (0, 1), (1, 0)),
                 'values': (0., 1., 10.),
                 'dense_shape': (2, 2)},
             'expected_input_layer': [
                 [[0.], [1.]],
                 [[10.], [0.]]],
             'expected_sequence_length': [2, 1]},
            {'testcase_name': '3D',
             'sparse_input_args': {
                 # feature 0, ids [[20, 3], [5]]
                 # feature 1, ids [[3], [8]]
                 'indices': ((0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 1, 0)),
                 'values': (20., 3., 5., 3., 8.),
                 'dense_shape': (2, 2, 2)},
             'expected_input_layer': [
                 [[20.], [3.], [5.], [0.]],
                 [[3.], [0.], [8.], [0.]]],
             'expected_sequence_length': [2, 2]},
        ]
        for param in params:
            self._numeric_column_multi_dim(
                sparse_input_args=param['sparse_input_args'],
                expected_input_layer=param['expected_input_layer'],
                expected_sequence_length=param['expected_sequence_length'],
            )

        params = [
            {'testcase_name': '2D',
             'sparse_input_args': {
                 # example 0, values [0., 1.,  2., 3., 4., 5., 6., 7.]
                 # example 1, [10., 11., 12., 13.]
                 'indices': ((0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6),
                             (0, 7), (1, 0), (1, 1), (1, 2), (1, 3)),
                 'values': (0., 1., 2., 3., 4., 5., 6., 7., 10., 11., 12., 13.),
                 'dense_shape': (2, 8)},
             'expected_input_layer': [
                 # The output of numeric_column._get_dense_tensor should be flattened.
                 [[0., 1., 2., 3.], [4., 5., 6., 7.]],
                 [[10., 11., 12., 13.], [0., 0., 0., 0.]]],
             'expected_sequence_length': [2, 1]},
            {'testcase_name': '3D',
             'sparse_input_args': {
                 # example 0, values [[0., 1., 2., 3.]], [[4., 5., 6., 7.]]
                 # example 1, [[10., 11., 12., 13.], []]
                 'indices': ((0, 0, 0), (0, 0, 1), (0, 0, 2), (0, 0, 3),
                             (0, 1, 0), (0, 1, 1), (0, 1, 2), (0, 1, 3),
                             (1, 0, 0), (1, 0, 1), (1, 0, 2), (1, 0, 3)),
                 'values': (0., 1., 2., 3., 4., 5., 6., 7., 10., 11., 12., 13.),
                 'dense_shape': (2, 2, 4)},
             'expected_input_layer': [
                 # The output of numeric_column._get_dense_tensor should be flattened.
                 [[0., 1., 2., 3.], [4., 5., 6., 7.]],
                 [[10., 11., 12., 13.], [0., 0., 0., 0.]]],
             'expected_sequence_length': [2, 1]},
        ]
        for param in params:
            self._numeric_column(
                sparse_input_args=param['sparse_input_args'],
                expected_input_layer=param['expected_input_layer'],
                expected_sequence_length=param['expected_sequence_length'],
            )

        params = [
            {'testcase_name': '2D',
             'sparse_input_args': {
                 # example 0, values [[[0., 1.],  [2., 3.]], [[4., 5.],  [6., 7.]]]
                 # example 1, [[[10., 11.],  [12., 13.]]]
                 'indices': ((0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6),
                             (0, 7), (1, 0), (1, 1), (1, 2), (1, 3)),
                 'values': (0., 1., 2., 3., 4., 5., 6., 7., 10., 11., 12., 13.),
                 'dense_shape': (2, 8)},
             'expected_shape': [2, 2, 4]},
            {'testcase_name': '3D',
             'sparse_input_args': {
                 # example 0, values [[0., 1., 2., 3.]], [[4., 5., 6., 7.]]
                 # example 1, [[10., 11., 12., 13.], []]
                 'indices': ((0, 0, 0), (0, 0, 1), (0, 0, 2), (0, 0, 3),
                             (0, 1, 0), (0, 1, 1), (0, 1, 2), (0, 1, 3),
                             (1, 0, 0), (1, 0, 1), (1, 0, 2), (1, 0, 3)),
                 'values': (0., 1., 2., 3., 4., 5., 6., 7., 10., 11., 12., 13.),
                 'dense_shape': (2, 2, 4)},
             'expected_shape': [2, 2, 4]},
        ]
        for param in params:
            self._static_shape_from_tensors_numeric(
                sparse_input_args=param['sparse_input_args'],
                expected_shape=param['expected_shape'],
            )

        params = [
            {'testcase_name': '2D',
             'sparse_input_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 # example 2, ids []
                 # example 3, ids [1]
                 'indices': ((0, 0), (1, 0), (1, 1), (3, 0)),
                 'values': (2, 0, 1, 1),
                 'dense_shape': (4, 2)},
             'expected_shape': [4, 2, 3]},
            {'testcase_name': '3D',
             'sparse_input_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 # example 2, ids []
                 # example 3, ids [[1], [0, 2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0),
                             (3, 0, 0), (3, 1, 0), (3, 1, 1)),
                 'values': (2, 0, 1, 2, 1, 0, 2),
                 'dense_shape': (4, 2, 2)},
             'expected_shape': [4, 2, 3]}
        ]
        for param in params:
            self._static_shape_from_tensors_indicator(
                sparse_input_args=param['sparse_input_args'],
                expected_shape=param['expected_shape'],
            )

    def _embedding_column(self, sparse_input_args_a, sparse_input_args_b, expected_input_layer,
                          expected_sequence_length):
        sparse_input_a = sparse_tensor.SparseTensorValue(**sparse_input_args_a)
        sparse_input_b = sparse_tensor.SparseTensorValue(**sparse_input_args_b)
        vocabulary_size = 3
        embedding_dimension_a = 2
        embedding_values_a = (
            (1., 2.),  # id 0
            (3., 4.),  # id 1
            (5., 6.)  # id 2
        )
        embedding_dimension_b = 3
        embedding_values_b = (
            (11., 12., 13.),  # id 0
            (14., 15., 16.),  # id 1
            (17., 18., 19.)  # id 2
        )

        def _get_initializer(embedding_dimension, embedding_values):
            def _initializer(shape, dtype, partition_info=None):
                test_case.assertAllEqual((vocabulary_size, embedding_dimension), shape)
                test_case.assertEqual(dtypes.float32, dtype)
                test_case.assertIsNone(partition_info)
                return embedding_values

            return _initializer

        categorical_column_a = sfc.sequence_categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size)
        embedding_column_a = fc.embedding_column(
            categorical_column_a,
            dimension=embedding_dimension_a,
            initializer=_get_initializer(embedding_dimension_a, embedding_values_a))
        categorical_column_b = sfc.sequence_categorical_column_with_identity(
            key='bbb', num_buckets=vocabulary_size)
        embedding_column_b = fc.embedding_column(
            categorical_column_b,
            dimension=embedding_dimension_b,
            initializer=_get_initializer(embedding_dimension_b, embedding_values_b))

        # Test that columns are reordered alphabetically.
        sequence_input_layer = sfc.SequenceFeatures(
            [embedding_column_b, embedding_column_a])
        input_layer, sequence_length = sequence_input_layer({
            'aaa': sparse_input_a, 'bbb': sparse_input_b, })

        test_case.evaluate(variables_lib.global_variables_initializer())
        weights = sequence_input_layer.weights
        test_case.assertCountEqual(
            ('sequence_features/aaa_embedding/embedding_weights:0',
             'sequence_features/bbb_embedding/embedding_weights:0'),
            tuple([v.name for v in weights]))
        test_case.assertAllEqual(embedding_values_a, test_case.evaluate(weights[0]))
        test_case.assertAllEqual(embedding_values_b, test_case.evaluate(weights[1]))
        test_case.assertAllEqual(expected_input_layer, test_case.evaluate(input_layer))
        test_case.assertAllEqual(expected_sequence_length, test_case.evaluate(sequence_length))

    def _embedding_column_with_non_sequence_categorical(self):
        """Tests that error is raised for non-sequence embedding column."""
        vocabulary_size = 3
        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, ids [2]
            # example 1, ids [0, 1]
            indices=((0, 0), (1, 0), (1, 1)),
            values=(2, 0, 1),
            dense_shape=(2, 2))

        categorical_column_a = fc.categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size)
        embedding_column_a = fc.embedding_column(
            categorical_column_a, dimension=2)

        with test_case.assertRaisesRegexp(
                ValueError,
                r'In embedding_column: aaa_embedding\. categorical_column must be of '
                r'type SequenceCategoricalColumn to use SequenceFeatures\.'):
            sequence_input_layer = sfc.SequenceFeatures([embedding_column_a])
            _, _ = sequence_input_layer({'aaa': sparse_input})

    def _shared_embedding_column(self):
        with ops.Graph().as_default():
            vocabulary_size = 3
            sparse_input_a = sparse_tensor.SparseTensorValue(
                # example 0, ids [2]
                # example 1, ids [0, 1]
                indices=((0, 0), (1, 0), (1, 1)),
                values=(2, 0, 1),
                dense_shape=(2, 2))
            sparse_input_b = sparse_tensor.SparseTensorValue(
                # example 0, ids [1]
                # example 1, ids [2, 0]
                indices=((0, 0), (1, 0), (1, 1)),
                values=(1, 2, 0),
                dense_shape=(2, 2))

            embedding_dimension = 2
            embedding_values = (
                (1., 2.),  # id 0
                (3., 4.),  # id 1
                (5., 6.)  # id 2
            )

            def _get_initializer(embedding_dimension, embedding_values):
                def _initializer(shape, dtype, partition_info=None):
                    test_case.assertAllEqual((vocabulary_size, embedding_dimension), shape)
                    test_case.assertEqual(dtypes.float32, dtype)
                    test_case.assertIsNone(partition_info)
                    return embedding_values

                return _initializer

            expected_input_layer = [
                # example 0, ids_a [2], ids_b [1]
                [[5., 6., 3., 4.], [0., 0., 0., 0.]],
                # example 1, ids_a [0, 1], ids_b [2, 0]
                [[1., 2., 5., 6.], [3., 4., 1., 2.]],
            ]
            expected_sequence_length = [1, 2]

            categorical_column_a = sfc.sequence_categorical_column_with_identity(
                key='aaa', num_buckets=vocabulary_size)
            categorical_column_b = sfc.sequence_categorical_column_with_identity(
                key='bbb', num_buckets=vocabulary_size)
            # Test that columns are reordered alphabetically.
            shared_embedding_columns = fc.shared_embedding_columns_v2(
                [categorical_column_b, categorical_column_a],
                dimension=embedding_dimension,
                initializer=_get_initializer(embedding_dimension, embedding_values))

            sequence_input_layer = sfc.SequenceFeatures(shared_embedding_columns)
            input_layer, sequence_length = sequence_input_layer({'aaa': sparse_input_a, 'bbb': sparse_input_b})

            global_vars = ops.get_collection(ops.GraphKeys.GLOBAL_VARIABLES)
            test_case.assertCountEqual(('aaa_bbb_shared_embedding:0',), tuple([v.name for v in global_vars]))
            with _initialized_session() as sess:
                test_case.assertAllEqual(embedding_values, global_vars[0].eval(session=sess))
                test_case.assertAllEqual(expected_input_layer, input_layer.eval(session=sess))
                test_case.assertAllEqual(expected_sequence_length, sequence_length.eval(session=sess))

    def _shared_embedding_column_with_non_sequence_categorical(self):
        """Tests that error is raised for non-sequence shared embedding column."""
        vocabulary_size = 3
        sparse_input_a = sparse_tensor.SparseTensorValue(
            # example 0, ids [2]
            # example 1, ids [0, 1]
            indices=((0, 0), (1, 0), (1, 1)),
            values=(2, 0, 1),
            dense_shape=(2, 2))
        sparse_input_b = sparse_tensor.SparseTensorValue(
            # example 0, ids [2]
            # example 1, ids [0, 1]
            indices=((0, 0), (1, 0), (1, 1)),
            values=(2, 0, 1),
            dense_shape=(2, 2))

        categorical_column_a = fc.categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        categorical_column_b = fc.categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)
        shared_embedding_columns = fc.shared_embedding_columns_v2([categorical_column_a, categorical_column_b],
                                                                  dimension=2)

        with test_case.assertRaisesRegexp(
                ValueError,
                r'In embedding_column: aaa_shared_embedding\. categorical_column must '
                r'be of type SequenceCategoricalColumn to use SequenceFeatures\.'):
            sequence_input_layer = sfc.SequenceFeatures(shared_embedding_columns)
            _, _ = sequence_input_layer({'aaa': sparse_input_a, 'bbb': sparse_input_b})

    def _indicator_column(self, sparse_input_args_a, sparse_input_args_b, expected_input_layer,
                          expected_sequence_length):
        sparse_input_a = sparse_tensor.SparseTensorValue(**sparse_input_args_a)
        sparse_input_b = sparse_tensor.SparseTensorValue(**sparse_input_args_b)

        vocabulary_size_a = 3
        vocabulary_size_b = 2

        categorical_column_a = sfc.sequence_categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size_a)
        indicator_column_a = fc.indicator_column(categorical_column_a)
        categorical_column_b = sfc.sequence_categorical_column_with_identity(
            key='bbb', num_buckets=vocabulary_size_b)
        indicator_column_b = fc.indicator_column(categorical_column_b)
        # Test that columns are reordered alphabetically.
        sequence_input_layer = sfc.SequenceFeatures(
            [indicator_column_b, indicator_column_a])
        input_layer, sequence_length = sequence_input_layer({
            'aaa': sparse_input_a, 'bbb': sparse_input_b})

        test_case.assertAllEqual(expected_input_layer, test_case.evaluate(input_layer))
        test_case.assertAllEqual(expected_sequence_length, test_case.evaluate(sequence_length))

    def _indicator_column_with_non_sequence_categorical(self):
        """Tests that error is raised for non-sequence categorical column."""
        vocabulary_size = 3
        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, ids [2]
            # example 1, ids [0, 1]
            indices=((0, 0), (1, 0), (1, 1)),
            values=(2, 0, 1),
            dense_shape=(2, 2))

        categorical_column_a = fc.categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size)
        indicator_column_a = fc.indicator_column(categorical_column_a)

        with test_case.assertRaisesRegexp(
                ValueError,
                r'In indicator_column: aaa_indicator\. categorical_column must be of '
                r'type SequenceCategoricalColumn to use SequenceFeatures\.'):
            sequence_input_layer = sfc.SequenceFeatures([indicator_column_a])
            _, _ = sequence_input_layer({'aaa': sparse_input})

    def _numeric_column(self, sparse_input_args, expected_input_layer, expected_sequence_length):
        sparse_input = sparse_tensor.SparseTensorValue(**sparse_input_args)

        numeric_column = sfc.sequence_numeric_column('aaa')

        sequence_input_layer = sfc.SequenceFeatures([numeric_column])
        input_layer, sequence_length = sequence_input_layer({'aaa': sparse_input})

        test_case.assertAllEqual(expected_input_layer, test_case.evaluate(input_layer))
        test_case.assertAllEqual(
            expected_sequence_length, test_case.evaluate(sequence_length))

    def _numeric_column_multi_dim(self, sparse_input_args, expected_input_layer, expected_sequence_length):
        """Tests SequenceFeatures for multi-dimensional numeric_column."""
        sparse_input = sparse_tensor.SparseTensorValue(**sparse_input_args)

        numeric_column = sfc.sequence_numeric_column('aaa', shape=(2, 2))

        sequence_input_layer = sfc.SequenceFeatures([numeric_column])
        input_layer, sequence_length = sequence_input_layer({'aaa': sparse_input})

        test_case.assertAllEqual(expected_input_layer, test_case.evaluate(input_layer))
        test_case.assertAllEqual(expected_sequence_length, test_case.evaluate(sequence_length))

    def _sequence_length_not_equal(self):
        """Tests that an error is raised when sequence lengths are not equal."""
        # Input a with sequence_length = [2, 1]
        sparse_input_a = sparse_tensor.SparseTensorValue(
            indices=((0, 0), (0, 1), (1, 0)),
            values=(0., 1., 10.),
            dense_shape=(2, 2))
        # Input b with sequence_length = [1, 1]
        sparse_input_b = sparse_tensor.SparseTensorValue(
            indices=((0, 0), (1, 0)),
            values=(1., 10.),
            dense_shape=(2, 2))
        numeric_column_a = sfc.sequence_numeric_column('aaa')
        numeric_column_b = sfc.sequence_numeric_column('bbb')

        sequence_input_layer = sfc.SequenceFeatures(
            [numeric_column_a, numeric_column_b])

        with test_case.assertRaisesRegexp(errors.InvalidArgumentError, r'Condition x == y did not hold.*'):
            _, sequence_length = sequence_input_layer({'aaa': sparse_input_a, 'bbb': sparse_input_b})
            test_case.evaluate(sequence_length)

    def _static_shape_from_tensors_numeric(self, sparse_input_args, expected_shape):
        """Tests that we return a known static shape when we have one."""
        sparse_input = sparse_tensor.SparseTensorValue(**sparse_input_args)
        numeric_column = sfc.sequence_numeric_column('aaa', shape=(2, 2))

        sequence_input_layer = sfc.SequenceFeatures([numeric_column])
        input_layer, _ = sequence_input_layer({'aaa': sparse_input})
        shape = input_layer.get_shape()
        test_case.assertEqual(shape, expected_shape)

    def _static_shape_from_tensors_indicator(self, sparse_input_args, expected_shape):
        """Tests that we return a known static shape when we have one."""
        sparse_input = sparse_tensor.SparseTensorValue(**sparse_input_args)
        categorical_column = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=3)
        indicator_column = fc.indicator_column(categorical_column)

        sequence_input_layer = sfc.SequenceFeatures([indicator_column])
        input_layer, _ = sequence_input_layer({'aaa': sparse_input})
        shape = input_layer.get_shape()
        test_case.assertEqual(shape, expected_shape)

    def _compute_output_shape(self):
        price1 = sfc.sequence_numeric_column('price1', shape=2)
        price2 = sfc.sequence_numeric_column('price2')
        features = {
            'price1': sparse_tensor.SparseTensor(
                indices=[[0, 0, 0], [0, 0, 1],
                         [0, 1, 0], [0, 1, 1],
                         [1, 0, 0], [1, 0, 1],
                         [2, 0, 0], [2, 0, 1],
                         [3, 0, 0], [3, 0, 1]],
                values=[0., 1., 10., 11., 100., 101., 200., 201., 300., 301.],
                dense_shape=(4, 3, 2)),
            'price2': sparse_tensor.SparseTensor(
                indices=[[0, 0],
                         [0, 1],
                         [1, 0],
                         [2, 0],
                         [3, 0]],
                values=[10., 11., 20., 30., 40.],
                dense_shape=(4, 3))}
        sequence_features = sfc.SequenceFeatures([price1, price2])
        seq_input, seq_len = sequence_features(features)
        test_case.assertEqual(sequence_features.compute_output_shape((None, None)), (None, None, 3))
        test_case.evaluate(variables_lib.global_variables_initializer())
        test_case.evaluate(lookup_ops.tables_initializer())

        test_case.assertAllClose([[[0., 1., 10.], [10., 11., 11.], [0., 0., 0.]],
                                  [[100., 101., 20.], [0., 0., 0.], [0., 0., 0.]],
                                  [[200., 201., 30.], [0., 0., 0.], [0., 0., 0.]],
                                  [[300., 301., 40.], [0., 0., 0.], [0., 0., 0.]]],
                                 test_case.evaluate(seq_input))
        test_case.assertAllClose([2, 1, 1, 1], test_case.evaluate(seq_len))


class ConcatenateContextInputTes:
    """Tests the utility fn concatenate_context_input."""

    def run(self):
        params = [
            {'testcase_name': 'rank_lt_3',
             'seq_input_arg': np.arange(100).reshape(10, 10)},
            {'testcase_name': 'rank_gt_3',
             'seq_input_arg': np.arange(100).reshape((5, 5, 2, 2))}
        ]
        for param in params:
            self._sequence_input_throws_error(seq_input_arg=param['seq_input_arg'])
        params = [
            {'testcase_name': 'rank_lt_2',
             'context_input_arg': np.arange(100)},
            {'testcase_name': 'rank_gt_2',
             'context_input_arg': np.arange(100).reshape((5, 5, 4))}
        ]
        for param in params:
            self._context_input_throws_error(context_input_arg=param['context_input_arg'])

    def _concatenate_context_input(self):
        seq_input = ops.convert_to_tensor(np.arange(12).reshape((2, 3, 2)))
        context_input = ops.convert_to_tensor(np.arange(10).reshape(2, 5))
        seq_input = math_ops.cast(seq_input, dtype=dtypes.float32)
        context_input = math_ops.cast(context_input, dtype=dtypes.float32)
        input_layer = sfc.concatenate_context_input(context_input, seq_input)

        expected = np.array([
            [[0, 1, 0, 1, 2, 3, 4], [2, 3, 0, 1, 2, 3, 4], [4, 5, 0, 1, 2, 3, 4]],
            [[6, 7, 5, 6, 7, 8, 9], [8, 9, 5, 6, 7, 8, 9], [10, 11, 5, 6, 7, 8, 9]]
        ], dtype=np.float32)
        output = test_case.evaluate(input_layer)
        test_case.assertAllEqual(expected, output)

    def _sequence_input_throws_error(self, seq_input_arg):
        seq_input = ops.convert_to_tensor(seq_input_arg)
        context_input = ops.convert_to_tensor(np.arange(100).reshape(10, 10))
        seq_input = math_ops.cast(seq_input, dtype=dtypes.float32)
        context_input = math_ops.cast(context_input, dtype=dtypes.float32)
        with test_case.assertRaisesRegexp(ValueError, 'sequence_input must have rank 3'):
            sfc.concatenate_context_input(context_input, seq_input)

    def _context_input_throws_error(self, context_input_arg):
        context_input = ops.convert_to_tensor(context_input_arg)
        seq_input = ops.convert_to_tensor(np.arange(100).reshape((5, 5, 4)))
        seq_input = math_ops.cast(seq_input, dtype=dtypes.float32)
        context_input = math_ops.cast(context_input, dtype=dtypes.float32)
        with test_case.assertRaisesRegexp(ValueError, 'context_input must have rank 2'):
            sfc.concatenate_context_input(context_input, seq_input)

    def _integer_seq_input_throws_error(self):
        seq_input = ops.convert_to_tensor(np.arange(100).reshape((5, 5, 4)))
        context_input = ops.convert_to_tensor(np.arange(100).reshape(10, 10))
        context_input = math_ops.cast(context_input, dtype=dtypes.float32)
        with test_case.assertRaisesRegexp(TypeError, 'sequence_input must have dtype float32'):
            sfc.concatenate_context_input(context_input, seq_input)

    def _integer_context_input_throws_error(self):
        seq_input = ops.convert_to_tensor(np.arange(100).reshape([5, 5, 4]))
        context_input = ops.convert_to_tensor(np.arange(100).reshape(10, 10))
        seq_input = math_ops.cast(seq_input, dtype=dtypes.float32)
        with test_case.assertRaisesRegexp(TypeError, 'context_input must have dtype float32'):
            sfc.concatenate_context_input(context_input, seq_input)


class DenseFeaturesTes:
    """Tests DenseFeatures with sequence feature columns."""

    def _get_categorical(self):
        """Tests that error is raised for sequence embedding column."""
        vocabulary_size = 3
        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, ids [2]
            # example 1, ids [0, 1]
            indices=((0, 0), (1, 0), (1, 1)),
            values=(2, 0, 1),
            dense_shape=(2, 2))

        categorical_column_a = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        return categorical_column_a, sparse_input

    def _embedding_column(self):
        """Tests that error is raised for sequence embedding column."""
        categorical_column_a, sparse_input = self._get_categorical()
        embedding_column_a = fc.embedding_column(categorical_column_a, dimension=2)

        with test_case.assertRaisesRegexp(
                ValueError,
                r'In embedding_column: aaa_embedding\. categorical_column must not be '
                r'of type SequenceCategoricalColumn\.'):
            input_layer = dense_features.DenseFeatures([embedding_column_a])
            _ = input_layer({'aaa': sparse_input})

    def _indicator_column(self):
        """Tests that error is raised for sequence indicator column."""
        categorical_column_a, sparse_input = self._get_categorical()
        indicator_column_a = fc.indicator_column(categorical_column_a)

        with test_case.assertRaisesRegexp(
                ValueError,
                r'In indicator_column: aaa_indicator\. categorical_column must not be '
                r'of type SequenceCategoricalColumn\.'):
            input_layer = dense_features.DenseFeatures([indicator_column_a])
            _ = input_layer({'aaa': sparse_input})

    def run(self):
        self._embedding_column()
        self._indicator_column()


class SequenceCategoricalColumnWithIdentityTes:
    def _get_sparse_tensors(self, inputs_args, expected_args):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        expected = sparse_tensor.SparseTensorValue(**expected_args)
        column = sfc.sequence_categorical_column_with_identity('aaa', num_buckets=9)

        id_weight_pair = _get_sparse_tensors(column, {'aaa': inputs})

        test_case.assertIsNone(id_weight_pair.weight_tensor)
        actual = test_case.evaluate(id_weight_pair.id_tensor)

        print(actual)
        _assert_sparse_tensor_value(expected, actual)

    def test_serialization(self):
        """Tests that column can be serialized."""
        parent = sfc.sequence_categorical_column_with_identity('animal', num_buckets=4)
        animal = fc.indicator_column(parent)

        config = animal.get_config()
        test_case.assertEqual(
            {
                'categorical_column': {
                    'class_name': 'SequenceCategoricalColumn',
                    'config': {
                        'categorical_column': {
                            'class_name': 'IdentityCategoricalColumn',
                            'config': {
                                'default_value': None,
                                'key': 'animal',
                                'number_buckets': 4
                            }
                        }
                    }
                }
            }, config)

        new_animal = fc.IndicatorColumn.from_config(config)
        test_case.assertEqual(animal, new_animal)
        test_case.assertIsNot(parent, new_animal.categorical_column)

        new_animal = fc.IndicatorColumn.from_config(config,
                                                    columns_by_name={
                                                        serialization._column_name_with_class_name(parent): parent
                                                    })
        test_case.assertEqual(animal, new_animal)
        test_case.assertIs(parent, new_animal.categorical_column)

    def run(self):
        # 2D
        self._get_sparse_tensors(
            inputs_args={
                'indices': ((0, 0), (1, 0), (1, 1)),
                'values': (1, 2, 0),
                'dense_shape': (2, 2)},
            expected_args={
                'indices': ((0, 0, 0), (1, 0, 0), (1, 1, 0)),
                'values': np.array((1, 2, 0), dtype=np.int64),
                'dense_shape': (2, 2, 1)})
        # 3D
        self._get_sparse_tensors(
            inputs_args={
                'indices': ((0, 0, 2), (1, 0, 0), (1, 2, 0)),
                'values': (6, 7, 8),
                'dense_shape': (2, 2, 2)},
            expected_args={
                'indices': ((0, 0, 2), (1, 0, 0), (1, 2, 0)),
                'values': np.array((6, 7, 8), dtype=np.int64),
                'dense_shape': (2, 2, 2)})


class SequenceCategoricalColumnWithHashBucketTes:

    def _get_sparse_tensors(self, inputs_args, expected_args):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        expected = sparse_tensor.SparseTensorValue(**expected_args)
        column = sfc.sequence_categorical_column_with_hash_bucket('aaa', hash_bucket_size=10)

        id_weight_pair = _get_sparse_tensors(column, {'aaa': inputs})

        test_case.assertIsNone(id_weight_pair.weight_tensor)
        actual = test_case.evaluate(id_weight_pair.id_tensor)
        _assert_sparse_tensor_indices_shape(expected, actual)

    def run(self):
        # 2D
        self._get_sparse_tensors(
            inputs_args={
                'indices': ((0, 0), (1, 0), (1, 1)),
                'values': ('omar', 'stringer', 'marlo'),
                'dense_shape': (2, 2)},
            expected_args={
                'indices': ((0, 0, 0), (1, 0, 0), (1, 1, 0)),
                # Ignored to avoid hash dependence in test.
                'values': np.array((0, 0, 0), dtype=np.int64),
                'dense_shape': (2, 2, 1)})
        # 3D
        self._get_sparse_tensors(
            inputs_args={
                'indices': ((0, 0, 2), (1, 0, 0), (1, 2, 0)),
                'values': ('omar', 'stringer', 'marlo'),
                'dense_shape': (2, 2, 2)},
            expected_args={
                'indices': ((0, 0, 2), (1, 0, 0), (1, 2, 0)),
                # Ignored to avoid hash dependence in test.
                'values': np.array((0, 0, 0), dtype=np.int64),
                'dense_shape': (2, 2, 2)})


class SequenceCategoricalColumnWithVocabularyListTes:

    def _get_sparse_tensors(self, inputs_args, expected_args):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        expected = sparse_tensor.SparseTensorValue(**expected_args)
        column = sfc.sequence_categorical_column_with_vocabulary_list(
            key='aaa',
            vocabulary_list=('omar', 'stringer', 'marlo'))

        id_weight_pair = _get_sparse_tensors(column, {'aaa': inputs})

        test_case.assertIsNone(id_weight_pair.weight_tensor)
        test_case.evaluate(variables_lib.global_variables_initializer())
        test_case.evaluate(lookup_ops.tables_initializer())
        actual = test_case.evaluate(id_weight_pair.id_tensor)
        _assert_sparse_tensor_value(expected, actual)

    def run(self):
        self._get_sparse_tensors(
            inputs_args={
                'indices': ((0, 0), (1, 0), (1, 1)),
                'values': ('marlo', 'skywalker', 'omar'),
                'dense_shape': (2, 2)},
            expected_args={
                'indices': ((0, 0, 0), (1, 0, 0), (1, 1, 0)),
                'values': np.array((2, -1, 0), dtype=np.int64),
                'dense_shape': (2, 2, 1)}
        )
        self._get_sparse_tensors(
            inputs_args={
                'indices': ((0, 0, 2), (1, 0, 0), (1, 2, 0)),
                'values': ('omar', 'skywalker', 'marlo'),
                'dense_shape': (2, 2, 2)},
            expected_args={
                'indices': ((0, 0, 2), (1, 0, 0), (1, 2, 0)),
                'values': np.array((0, -1, 2), dtype=np.int64),
                'dense_shape': (2, 2, 2)})


class SequenceEmbeddingColumnTes:

    def _get_sequence_dense_tensor(self, inputs_args, expected):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        vocabulary_size = 3
        embedding_dimension = 2

        categorical_column = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        embedding_column = fc.embedding_column(categorical_column, dimension=embedding_dimension, )
        embedding_lookup, _, state_manager = _get_sequence_dense_tensor_state(embedding_column, {'aaa': inputs})

        embedding_column2 = fc.embedding_column(categorical_column, dimension=embedding_dimension, )
        embedding_lookup2, _, state_manager = _get_sequence_dense_tensor_state(embedding_column2, {'aaa': inputs})

        actual = test_case.evaluate(embedding_lookup)
        actual2 = test_case.evaluate(embedding_lookup2)

        print(actual)
        print(actual2)
        print("\n\n\n\n")

    def _sequence_length(self, inputs_args, expected_sequence_length):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        vocabulary_size = 3

        categorical_column = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        embedding_column = fc.embedding_column(categorical_column, dimension=2)

        _, sequence_length, _ = _get_sequence_dense_tensor_state(embedding_column, {'aaa': inputs})

        sequence_length = test_case.evaluate(sequence_length)
        test_case.assertAllEqual(expected_sequence_length, sequence_length)
        test_case.assertEqual(np.int64, sequence_length.dtype)

    def _sequence_length_with_empty_rows(self):
        """Tests _sequence_length when some examples do not have ids."""
        vocabulary_size = 3
        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, ids []
            # example 1, ids [2]
            # example 2, ids [0, 1]
            # example 3, ids []
            # example 4, ids [1]
            # example 5, ids []
            indices=((1, 0), (2, 0), (2, 1), (4, 0)),
            values=(2, 0, 1, 1),
            dense_shape=(6, 2))
        expected_sequence_length = [0, 1, 2, 0, 1, 0]

        categorical_column = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        embedding_column = fc.embedding_column(categorical_column, dimension=2)

        _, sequence_length, _ = _get_sequence_dense_tensor_state(embedding_column, {'aaa': sparse_input})

        test_case.assertAllEqual(expected_sequence_length, test_case.evaluate(sequence_length))

    def run(self):
        params = [
            {'testcase_name': '2D',
             'inputs_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 # example 2, ids []
                 # example 3, ids [1]
                 'indices': ((0, 0), (1, 0), (1, 1), (3, 0)),
                 'values': (2, 0, 1, 1),
                 'dense_shape': (4, 2)},
             'expected': [
                 # example 0, ids [2]
                 [[7., 11.], [0., 0.]],
                 # example 1, ids [0, 1]
                 [[1., 2.], [3., 5.]],
                 # example 2, ids []
                 [[0., 0.], [0., 0.]],
                 # example 3, ids [1]
                 [[3., 5.], [0., 0.]]]},
            {'testcase_name': '3D',
             'inputs_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 # example 2, ids []
                 # example 3, ids [[1], [0, 2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0),
                             (3, 0, 0), (3, 1, 0), (3, 1, 1)),
                 'values': (2, 0, 1, 2, 1, 0, 2),
                 'dense_shape': (4, 2, 2)},
             'expected': [
                 # example 0, ids [[2]]
                 [[7., 11.], [0., 0.]],
                 # example 1, ids [[0, 1], [2]]
                 [[2, 3.5], [7., 11.]],
                 # example 2, ids []
                 [[0., 0.], [0., 0.]],
                 # example 3, ids [[1], [0, 2]]
                 [[3., 5.], [4., 6.5]]]}
        ]
        for param in params:
            self._get_sequence_dense_tensor(inputs_args=param['inputs_args'], expected=param['expected'])
        params = [
            {'testcase_name': '2D',
             'inputs_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (2, 0, 1),
                 'dense_shape': (2, 2)},
             'expected_sequence_length': [1, 2]},
            {'testcase_name': '3D',
             'inputs_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0)),
                 'values': (2, 0, 1, 2),
                 'dense_shape': (2, 2, 2)},
             'expected_sequence_length': [1, 2]}]

        for param in params:
            self._sequence_length(inputs_args=param['inputs_args'],
                                  expected_sequence_length=param['expected_sequence_length'])


class SequenceSharedEmbeddingColumnTes:

    def _get_sequence_dense_tensor(self):
        vocabulary_size = 3
        embedding_dimension = 2
        embedding_values = (
            (1., 2.),  # id 0
            (3., 5.),  # id 1
            (7., 11.)  # id 2
        )

        def _initializer(shape, dtype, partition_info=None):
            test_case.assertAllEqual((vocabulary_size, embedding_dimension), shape)
            test_case.assertEqual(dtypes.float32, dtype)
            test_case.assertIsNone(partition_info)
            return embedding_values

        sparse_input_a = sparse_tensor.SparseTensorValue(
            # example 0, ids [2]
            # example 1, ids [0, 1]
            # example 2, ids []
            # example 3, ids [1]
            indices=((0, 0), (1, 0), (1, 1), (3, 0)),
            values=(2, 0, 1, 1),
            dense_shape=(4, 2))
        sparse_input_b = sparse_tensor.SparseTensorValue(
            # example 0, ids [1]
            # example 1, ids [0, 2]
            # example 2, ids [0]
            # example 3, ids []
            indices=((0, 0), (1, 0), (1, 1), (2, 0)),
            values=(1, 0, 2, 0),
            dense_shape=(4, 2))

        expected_lookups_a = [
            # example 0, ids [2]
            [[7., 11.], [0., 0.]],
            # example 1, ids [0, 1]
            [[1., 2.], [3., 5.]],
            # example 2, ids []
            [[0., 0.], [0., 0.]],
            # example 3, ids [1]
            [[3., 5.], [0., 0.]],
        ]

        expected_lookups_b = [
            # example 0, ids [1]
            [[3., 5.], [0., 0.]],
            # example 1, ids [0, 2]
            [[1., 2.], [7., 11.]],
            # example 2, ids [0]
            [[1., 2.], [0., 0.]],
            # example 3, ids []
            [[0., 0.], [0., 0.]],
        ]

        categorical_column_a = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        categorical_column_b = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)
        shared_embedding_columns = fc.shared_embedding_columns_v2(
            [categorical_column_a, categorical_column_b],
            dimension=embedding_dimension,
            initializer=_initializer)

        embedding_lookup_a = _get_sequence_dense_tensor(shared_embedding_columns[0], {'aaa': sparse_input_a})[0]
        embedding_lookup_b = _get_sequence_dense_tensor(shared_embedding_columns[1], {'bbb': sparse_input_b})[0]

        test_case.evaluate(variables_lib.global_variables_initializer())
        global_vars = ops.get_collection(ops.GraphKeys.GLOBAL_VARIABLES)
        test_case.assertItemsEqual(('aaa_bbb_shared_embedding:0',), tuple([v.name for v in global_vars]))
        test_case.assertAllEqual(embedding_values, test_case.evaluate(global_vars[0]))
        test_case.assertAllEqual(expected_lookups_a, test_case.evaluate(embedding_lookup_a))
        test_case.assertAllEqual(expected_lookups_b, test_case.evaluate(embedding_lookup_b))

    def _sequence_length(self):
        with ops.Graph().as_default():
            vocabulary_size = 3

            sparse_input_a = sparse_tensor.SparseTensorValue(
                # example 0, ids [2]
                # example 1, ids [0, 1]
                indices=((0, 0), (1, 0), (1, 1)),
                values=(2, 0, 1),
                dense_shape=(2, 2))
            expected_sequence_length_a = [1, 2]
            categorical_column_a = sfc.sequence_categorical_column_with_identity(
                key='aaa', num_buckets=vocabulary_size)

            sparse_input_b = sparse_tensor.SparseTensorValue(
                # example 0, ids [0, 2]
                # example 1, ids [1]
                indices=((0, 0), (0, 1), (1, 0)),
                values=(0, 2, 1),
                dense_shape=(2, 2))
            expected_sequence_length_b = [2, 1]
            categorical_column_b = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)
            shared_embedding_columns = fc.shared_embedding_columns_v2([categorical_column_a, categorical_column_b],
                                                                      dimension=2)

            sequence_length_a = _get_sequence_dense_tensor(shared_embedding_columns[0], {'aaa': sparse_input_a})[1]
            sequence_length_b = _get_sequence_dense_tensor(shared_embedding_columns[1], {'bbb': sparse_input_b})[1]

            with _initialized_session() as sess:
                sequence_length_a = sess.run(sequence_length_a)
                test_case.assertAllEqual(expected_sequence_length_a, sequence_length_a)
                test_case.assertEqual(np.int64, sequence_length_a.dtype)
                sequence_length_b = sess.run(sequence_length_b)
                test_case.assertAllEqual(expected_sequence_length_b, sequence_length_b)
                test_case.assertEqual(np.int64, sequence_length_b.dtype)

    def _sequence_length_with_empty_rows(self):
        """Tests _sequence_length when some examples do not have ids."""
        with ops.Graph().as_default():
            vocabulary_size = 3
            sparse_input_a = sparse_tensor.SparseTensorValue(
                # example 0, ids []
                # example 1, ids [2]
                # example 2, ids [0, 1]
                # example 3, ids []
                # example 4, ids [1]
                # example 5, ids []
                indices=((1, 0), (2, 0), (2, 1), (4, 0)),
                values=(2, 0, 1, 1),
                dense_shape=(6, 2))
            expected_sequence_length_a = [0, 1, 2, 0, 1, 0]
            categorical_column_a = sfc.sequence_categorical_column_with_identity(
                key='aaa', num_buckets=vocabulary_size)

            sparse_input_b = sparse_tensor.SparseTensorValue(
                # example 0, ids [2]
                # example 1, ids []
                # example 2, ids []
                # example 3, ids []
                # example 4, ids [1]
                # example 5, ids [0, 1]
                indices=((0, 0), (4, 0), (5, 0), (5, 1)),
                values=(2, 1, 0, 1),
                dense_shape=(6, 2))
            expected_sequence_length_b = [1, 0, 0, 0, 1, 2]
            categorical_column_b = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)

            shared_embedding_columns = fc.shared_embedding_columns_v2([categorical_column_a, categorical_column_b],
                                                                      dimension=2)

            sequence_length_a = _get_sequence_dense_tensor(shared_embedding_columns[0], {'aaa': sparse_input_a})[1]
            sequence_length_b = _get_sequence_dense_tensor(shared_embedding_columns[1], {'bbb': sparse_input_b})[1]

            with _initialized_session() as sess:
                test_case.assertAllEqual(expected_sequence_length_a, sequence_length_a.eval(session=sess))
                test_case.assertAllEqual(expected_sequence_length_b, sequence_length_b.eval(session=sess))

    def run(self):
        self._get_sequence_dense_tensor()
        self._sequence_length()
        self._sequence_length_with_empty_rows()


class SequenceSharedEmbeddingColumnTes2:

    def _get_sequence_dense_tensor(self):
        vocabulary_size = 3
        embedding_dimension = 2

        sparse_input_a = sparse_tensor.SparseTensorValue(
            indices=((0, 0), (1, 0), (1, 1), (3, 0)),
            values=(2, 0, 1, 1),
            dense_shape=(4, 2))
        sparse_input_b = sparse_tensor.SparseTensorValue(
            indices=((0, 0), (1, 0), (1, 1), (2, 0)),
            values=(1, 0, 2, 0),
            dense_shape=(4, 2))

        categorical_column_a = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        categorical_column_b = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)
        categorical_column_c = sfc.sequence_categorical_column_with_identity(key='aaa', num_buckets=vocabulary_size)
        categorical_column_d = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)
        with tf.compat.v1.variable_scope("share", reuse=tf.compat.v1.AUTO_REUSE):
            shared_embedding_columns = fc.shared_embedding_columns_v2([categorical_column_a],
                                                                      dimension=embedding_dimension,
                                                                      shared_embedding_collection_name="a")
            shared_embedding_columns2 = fc.shared_embedding_columns_v2([categorical_column_c, categorical_column_d],
                                                                       dimension=embedding_dimension,
                                                                       shared_embedding_collection_name='a')

            embedding_lookup_a = sfc.SequenceFeatures(shared_embedding_columns[0])({'aaa': sparse_input_a})[0]
            # embedding_lookup_b = sfc.SequenceFeatures(shared_embedding_columns[1])({'bbb': sparse_input_b})[0]
            embedding_lookup_c = sfc.SequenceFeatures(shared_embedding_columns2[0])({'aaa': sparse_input_a})[0]
            embedding_lookup_d = _get_sequence_dense_tensor(shared_embedding_columns2[1], {'bbb': sparse_input_b})[0]

        test_case.evaluate(variables_lib.global_variables_initializer())

        actual1 = test_case.evaluate(embedding_lookup_a)
        actual2 = test_case.evaluate(embedding_lookup_c)

        print(actual1)
        print(actual2)

    def _sequence_length(self):
        with ops.Graph().as_default():
            vocabulary_size = 3

            sparse_input_a = sparse_tensor.SparseTensorValue(
                # example 0, ids [2]
                # example 1, ids [0, 1]
                indices=((0, 0), (1, 0), (1, 1)),
                values=(2, 0, 1),
                dense_shape=(2, 2))
            expected_sequence_length_a = [1, 2]
            categorical_column_a = sfc.sequence_categorical_column_with_identity(
                key='aaa', num_buckets=vocabulary_size)

            sparse_input_b = sparse_tensor.SparseTensorValue(
                # example 0, ids [0, 2]
                # example 1, ids [1]
                indices=((0, 0), (0, 1), (1, 0)),
                values=(0, 2, 1),
                dense_shape=(2, 2))
            expected_sequence_length_b = [2, 1]
            categorical_column_b = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)
            shared_embedding_columns = fc.shared_embedding_columns_v2([categorical_column_a, categorical_column_b],
                                                                      dimension=2)

            sequence_length_a = _get_sequence_dense_tensor(shared_embedding_columns[0], {'aaa': sparse_input_a})[1]
            sequence_length_b = _get_sequence_dense_tensor(shared_embedding_columns[1], {'bbb': sparse_input_b})[1]

            with _initialized_session() as sess:
                sequence_length_a = sess.run(sequence_length_a)
                test_case.assertAllEqual(expected_sequence_length_a, sequence_length_a)
                test_case.assertEqual(np.int64, sequence_length_a.dtype)
                sequence_length_b = sess.run(sequence_length_b)
                test_case.assertAllEqual(expected_sequence_length_b, sequence_length_b)
                test_case.assertEqual(np.int64, sequence_length_b.dtype)

    def _sequence_length_with_empty_rows(self):
        """Tests _sequence_length when some examples do not have ids."""
        with ops.Graph().as_default():
            vocabulary_size = 3
            sparse_input_a = sparse_tensor.SparseTensorValue(
                # example 0, ids []
                # example 1, ids [2]
                # example 2, ids [0, 1]
                # example 3, ids []
                # example 4, ids [1]
                # example 5, ids []
                indices=((1, 0), (2, 0), (2, 1), (4, 0)),
                values=(2, 0, 1, 1),
                dense_shape=(6, 2))
            expected_sequence_length_a = [0, 1, 2, 0, 1, 0]
            categorical_column_a = sfc.sequence_categorical_column_with_identity(
                key='aaa', num_buckets=vocabulary_size)

            sparse_input_b = sparse_tensor.SparseTensorValue(
                # example 0, ids [2]
                # example 1, ids []
                # example 2, ids []
                # example 3, ids []
                # example 4, ids [1]
                # example 5, ids [0, 1]
                indices=((0, 0), (4, 0), (5, 0), (5, 1)),
                values=(2, 1, 0, 1),
                dense_shape=(6, 2))
            expected_sequence_length_b = [1, 0, 0, 0, 1, 2]
            categorical_column_b = sfc.sequence_categorical_column_with_identity(key='bbb', num_buckets=vocabulary_size)

            shared_embedding_columns = fc.shared_embedding_columns_v2([categorical_column_a, categorical_column_b],
                                                                      dimension=2)

            sequence_length_a = _get_sequence_dense_tensor(shared_embedding_columns[0], {'aaa': sparse_input_a})[1]
            sequence_length_b = _get_sequence_dense_tensor(shared_embedding_columns[1], {'bbb': sparse_input_b})[1]

            with _initialized_session() as sess:
                test_case.assertAllEqual(expected_sequence_length_a, sequence_length_a.eval(session=sess))
                test_case.assertAllEqual(expected_sequence_length_b, sequence_length_b.eval(session=sess))

    def run(self):
        self._get_sequence_dense_tensor()
        self._sequence_length()
        self._sequence_length_with_empty_rows()


class SequenceIndicatorColumnTes:
    def run(self):
        params = [
            {'testcase_name': '2D',
             'inputs_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 # example 2, ids []
                 # example 3, ids [1]
                 'indices': ((0, 0), (1, 0), (1, 1), (3, 0)),
                 'values': (2, 0, 1, 1),
                 'dense_shape': (4, 2)},
             'expected': [
                 # example 0, ids [2]
                 [[0., 0., 1.], [0., 0., 0.]],
                 # example 1, ids [0, 1]
                 [[1., 0., 0.], [0., 1., 0.]],
                 # example 2, ids []
                 [[0., 0., 0.], [0., 0., 0.]],
                 # example 3, ids [1]
                 [[0., 1., 0.], [0., 0., 0.]]]},
            {'testcase_name': '3D',
             'inputs_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 # example 2, ids []
                 # example 3, ids [[1], [2, 2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0),
                             (3, 0, 0), (3, 1, 0), (3, 1, 1)),
                 'values': (2, 0, 1, 2, 1, 2, 2),
                 'dense_shape': (4, 2, 2)},
             'expected': [
                 # example 0, ids [[2]]
                 [[0., 0., 1.], [0., 0., 0.]],
                 # example 1, ids [[0, 1], [2]]
                 [[1., 1., 0.], [0., 0., 1.]],
                 # example 2, ids []
                 [[0., 0., 0.], [0., 0., 0.]],
                 # example 3, ids [[1], [2, 2]]
                 [[0., 1., 0.], [0., 0., 2.]]]}
        ]

        for param in params:
            self._get_sequence_dense_tensor(inputs_args=param['inputs_args'], expected=param['expected'])

        params = [
            {'testcase_name': '2D',
             'inputs_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (2, 0, 1),
                 'dense_shape': (2, 2)},
             'expected_sequence_length': [1, 2]},
            {'testcase_name': '3D',
             'inputs_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0)),
                 'values': (2, 0, 1, 2),
                 'dense_shape': (2, 2, 2)},
             'expected_sequence_length': [1, 2]}
        ]
        for param in params:
            self._sequence_length(inputs_args=param['inputs_args'],
                                  expected_sequence_length=param['expected_sequence_length'])

    def _get_sequence_dense_tensor(self, inputs_args, expected):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        vocabulary_size = 3

        categorical_column = sfc.sequence_categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size)
        indicator_column = fc.indicator_column(categorical_column)

        indicator_tensor, _ = _get_sequence_dense_tensor(
            indicator_column, {'aaa': inputs})

        test_case.assertAllEqual(expected, test_case.evaluate(indicator_tensor))

    def _sequence_length(self, inputs_args, expected_sequence_length):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        vocabulary_size = 3

        categorical_column = sfc.sequence_categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size)
        indicator_column = fc.indicator_column(categorical_column)

        _, sequence_length = _get_sequence_dense_tensor(
            indicator_column, {'aaa': inputs})

        sequence_length = test_case.evaluate(sequence_length)
        test_case.assertAllEqual(expected_sequence_length, sequence_length)
        test_case.assertEqual(np.int64, sequence_length.dtype)

    def _sequence_length_with_empty_rows(self):
        """Tests _sequence_length when some examples do not have ids."""
        vocabulary_size = 3
        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, ids []
            # example 1, ids [2]
            # example 2, ids [0, 1]
            # example 3, ids []
            # example 4, ids [1]
            # example 5, ids []
            indices=((1, 0), (2, 0), (2, 1), (4, 0)),
            values=(2, 0, 1, 1),
            dense_shape=(6, 2))
        expected_sequence_length = [0, 1, 2, 0, 1, 0]

        categorical_column = sfc.sequence_categorical_column_with_identity(
            key='aaa', num_buckets=vocabulary_size)
        indicator_column = fc.indicator_column(categorical_column)

        _, sequence_length = _get_sequence_dense_tensor(
            indicator_column, {'aaa': sparse_input})

        test_case.assertAllEqual(expected_sequence_length, test_case.evaluate(sequence_length))


class SequenceNumericColumnTes:

    def run(self):
        params = [
            {'testcase_name': '2D',
             'inputs_args': {
                 # example 0, values [0., 1]
                 # example 1, [10.]
                 'indices': ((0, 0), (0, 1), (1, 0)),
                 'values': (0., 1., 10.),
                 'dense_shape': (2, 2)},
             'expected': [
                 [[0.], [1.]],
                 [[10.], [0.]]]},
            {'testcase_name': '3D',
             'inputs_args': {
                 # feature 0, ids [[20, 3], [5]]
                 # feature 1, ids [[3], [8]]
                 'indices': ((0, 0, 0), (0, 0, 1), (0, 1, 0), (1, 0, 0), (1, 1, 0)),
                 'values': (20, 3, 5., 3., 8.),
                 'dense_shape': (2, 2, 2)},
             'expected': [
                 [[20.], [3.], [5.], [0.]],
                 [[3.], [0.], [8.], [0.]]]},
        ]

        for param in params:
            self._get_sequence_dense_tensor(inputs_args=param['inputs_args'], expected=param['expected'])

        params = [
            {'testcase_name': '2D',
             'sparse_input_args': {
                 # example 0, values [[[0., 1.],  [2., 3.]], [[4., 5.],  [6., 7.]]]
                 # example 1, [[[10., 11.],  [12., 13.]]]
                 'indices': ((0, 0), (0, 1), (0, 2), (0, 3), (0, 4), (0, 5), (0, 6),
                             (0, 7), (1, 0), (1, 1), (1, 2), (1, 3)),
                 'values': (0., 1., 2., 3., 4., 5., 6., 7., 10., 11., 12., 13.),
                 'dense_shape': (2, 8)},
             'expected_dense_tensor': [
                 [[[0., 1.], [2., 3.]], [[4., 5.], [6., 7.]]],
                 [[[10., 11.], [12., 13.]], [[0., 0.], [0., 0.]]]]},
            {'testcase_name': '3D',
             'sparse_input_args': {
                 'indices': ((0, 0, 0), (0, 0, 2), (0, 0, 4), (0, 0, 6),
                             (0, 1, 0), (0, 1, 2), (0, 1, 4), (0, 1, 6),
                             (1, 0, 0), (1, 0, 2), (1, 0, 4), (1, 0, 6)),
                 'values': (0., 1., 2., 3., 4., 5., 6., 7., 10., 11., 12., 13.),
                 'dense_shape': (2, 2, 8)},
             'expected_dense_tensor': [
                 [[[0., 0.], [1., 0.]], [[2., 0.], [3., 0.]],
                  [[4., 0.], [5., 0.]], [[6., 0.], [7., 0.]]],
                 [[[10., 0.], [11., 0.]], [[12., 0.], [13., 0.]],
                  [[0., 0.], [0., 0.]], [[0., 0.], [0., 0.]]]]},
        ]
        for param in params:
            self._get_dense_tensor_multi_dim(sparse_input_args=param['sparse_input_args'],
                                             expected_dense_tensor=param['expected_dense_tensor'])

        params = [
            {'testcase_name': '2D',
             'inputs_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (2., 0., 1.),
                 'dense_shape': (2, 2)},
             'expected_sequence_length': [1, 2],
             'shape': (1,)},
            {'testcase_name': '3D',
             'inputs_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0)),
                 'values': (2., 0., 1., 2.),
                 'dense_shape': (2, 2, 2)},
             'expected_sequence_length': [1, 2],
             'shape': (1,)},
            {'testcase_name': '2D_with_shape',
             'inputs_args': {
                 # example 0, ids [2]
                 # example 1, ids [0, 1]
                 'indices': ((0, 0), (1, 0), (1, 1)),
                 'values': (2., 0., 1.),
                 'dense_shape': (2, 2)},
             'expected_sequence_length': [1, 1],
             'shape': (2,)},
            {'testcase_name': '3D_with_shape',
             'inputs_args': {
                 # example 0, ids [[2]]
                 # example 1, ids [[0, 1], [2]]
                 'indices': ((0, 0, 0), (1, 0, 0), (1, 0, 1), (1, 1, 0)),
                 'values': (2., 0., 1., 2.),
                 'dense_shape': (2, 2, 2)},
             'expected_sequence_length': [1, 2],
             'shape': (2,)},
        ]
        for param in params:
            self._sequence_length(inputs_args=param['inputs_args'],
                                  expected_sequence_length=param['expected_sequence_length'], shape=param['shape'])

    def test_defaults(self):
        a = sfc.sequence_numeric_column('aaa')
        test_case.assertEqual('aaa', a.key)
        test_case.assertEqual('aaa', a.name)
        test_case.assertEqual((1,), a.shape)
        test_case.assertEqual(0., a.default_value)
        test_case.assertEqual(dtypes.float32, a.dtype)
        test_case.assertIsNone(a.normalizer_fn)

    def test_shape_saved_as_tuple(self):
        a = sfc.sequence_numeric_column('aaa', shape=[1, 2])
        test_case.assertEqual((1, 2), a.shape)

    def test_shape_must_be_positive_integer(self):
        with test_case.assertRaisesRegexp(TypeError, 'shape dimensions must be integer'):
            sfc.sequence_numeric_column('aaa', shape=[1.0])

        with test_case.assertRaisesRegexp(ValueError, 'shape dimensions must be greater than 0'):
            sfc.sequence_numeric_column('aaa', shape=[0])

    def test_dtype_is_convertible_to_float(self):
        with test_case.assertRaisesRegexp(ValueError, 'dtype must be convertible to float'):
            sfc.sequence_numeric_column('aaa', dtype=dtypes.string)

    def test_normalizer_fn_must_be_callable(self):
        with test_case.assertRaisesRegexp(TypeError, 'must be a callable'):
            sfc.sequence_numeric_column('aaa', normalizer_fn='NotACallable')

    def _get_sequence_dense_tensor(self, inputs_args, expected):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        numeric_column = sfc.sequence_numeric_column('aaa')

        dense_tensor, _ = _get_sequence_dense_tensor(
            numeric_column, {'aaa': inputs})
        test_case.assertAllEqual(expected, test_case.evaluate(dense_tensor))

    def _get_sequence_dense_tensor_with_normalizer_fn(self):
        def _increment_two(input_sparse_tensor):
            return sparse_ops.sparse_add(
                input_sparse_tensor,
                sparse_tensor.SparseTensor(((0, 0), (1, 1)), (2.0, 2.0), (2, 2))
            )

        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, values [[0.], [1]]
            # example 1, [[10.]]
            indices=((0, 0), (0, 1), (1, 0)),
            values=(0., 1., 10.),
            dense_shape=(2, 2))

        # Before _increment_two:
        #   [[0.], [1.]],
        #   [[10.], [0.]],
        # After _increment_two:
        #   [[2.], [1.]],
        #   [[10.], [2.]],
        expected_dense_tensor = [
            [[2.], [1.]],
            [[10.], [2.]],
        ]
        numeric_column = sfc.sequence_numeric_column(
            'aaa', normalizer_fn=_increment_two)

        dense_tensor, _ = _get_sequence_dense_tensor(
            numeric_column, {'aaa': sparse_input})

        test_case.assertAllEqual(expected_dense_tensor, test_case.evaluate(dense_tensor))

    def _get_dense_tensor_multi_dim(self, sparse_input_args, expected_dense_tensor):
        """Tests get_sequence_dense_tensor for multi-dim numeric_column."""
        sparse_input = sparse_tensor.SparseTensorValue(**sparse_input_args)
        numeric_column = sfc.sequence_numeric_column('aaa', shape=(2, 2))

        dense_tensor, _ = _get_sequence_dense_tensor(
            numeric_column, {'aaa': sparse_input})

        test_case.assertAllEqual(expected_dense_tensor, test_case.evaluate(dense_tensor))

    def _sequence_length(self, inputs_args, expected_sequence_length, shape):
        inputs = sparse_tensor.SparseTensorValue(**inputs_args)
        numeric_column = sfc.sequence_numeric_column('aaa', shape=shape)

        _, sequence_length = _get_sequence_dense_tensor(
            numeric_column, {'aaa': inputs})

        sequence_length = test_case.evaluate(sequence_length)
        test_case.assertAllEqual(expected_sequence_length, sequence_length)
        test_case.assertEqual(np.int64, sequence_length.dtype)

    def test_sequence_length_with_empty_rows(self):
        """Tests _sequence_length when some examples do not have ids."""
        sparse_input = sparse_tensor.SparseTensorValue(
            # example 0, values []
            # example 1, values [[0.], [1.]]
            # example 2, [[2.]]
            # example 3, values []
            # example 4, [[3.]]
            # example 5, values []
            indices=((1, 0), (1, 1), (2, 0), (4, 0)),
            values=(0., 1., 2., 3.),
            dense_shape=(6, 2))
        expected_sequence_length = [0, 2, 1, 0, 1, 0]
        numeric_column = sfc.sequence_numeric_column('aaa')

        _, sequence_length = _get_sequence_dense_tensor(
            numeric_column, {'aaa': sparse_input})

        test_case.assertAllEqual(expected_sequence_length, test_case.evaluate(sequence_length))

    def test_serialization(self):
        """Tests that column can be serialized."""

        def _custom_fn(input_tensor):
            return input_tensor + 42

        column = sfc.sequence_numeric_column(
            key='my-key', shape=(2,), default_value=3, dtype=dtypes.int32,
            normalizer_fn=_custom_fn)
        configs = serialization.serialize_feature_column(column)
        column = serialization.deserialize_feature_column(
            configs, custom_objects={_custom_fn.__name__: _custom_fn})
        test_case.assertEqual(column.key, 'my-key')
        test_case.assertEqual(column.shape, (2,))
        test_case.assertEqual(column.default_value, 3)
        test_case.assertEqual(column.normalizer_fn(3), 45)
        with test_case.assertRaisesRegex(ValueError, 'Instance: 0 is not a FeatureColumn'):
            serialization.serialize_feature_column(int())

    def test_parents(self):
        """Tests parents attribute of column."""
        column = sfc.sequence_numeric_column(key='my-key')
        test_case.assertEqual(column.parents, ['my-key'])


def example():
    URL = 'https://storage.googleapis.com/applied-dl/heart.csv'
    dataframe = pd.read_csv(URL)
    dataframe.head()

    train, test = train_test_split(dataframe, test_size=0.2)
    train, val = train_test_split(train, test_size=0.2)
    print(len(train), 'train examples')
    print(len(val), 'validation examples')
    print(len(test), 'test examples')

    import random
    list = ["a", 'b', 'c', 'd', 'e', 'f']

    def arr_c(n):
        res = []
        for i in range(0, n):
            random.shuffle(list)
            res.append(np.array(list[:4]))
        return res

    # 一种从 Pandas Dataframe 创建 tf.data 数据集的实用程序方法（utility method）
    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        labels = dataframe.pop('target')
        dataframe['arr'] = arr_c(len(dataframe))
        ds = tf.data.Dataset.from_tensor_slices((dict(dataframe), labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds

    batch_size = 5  # 小批量大小用于演示
    train_ds = df_to_dataset(train, batch_size=batch_size)
    val_ds = df_to_dataset(val, shuffle=False, batch_size=batch_size)
    test_ds = df_to_dataset(test, shuffle=False, batch_size=batch_size)

    # 我们将使用该批数据演示几种特征列
    example_batch = next(iter(train_ds))[0]

    from tensorflow.python.feature_column import sequence_feature_column as sfc

    thal_seq = feature_column.sequence_categorical_column_with_vocabulary_list('arr', ['a', 'b', 'c', 'd', 'e', 'f'])

    # thal_one_hot = feature_column.embedding_column(thal_seq, dimension=8)
    # thal_one_hot = feature_column.indicator_column(thal_seq)
    from notekeras.feature.feature_column_def import IndicatorColumnDef

    thal_one_hot = IndicatorColumnDef(thal_seq, size=15)

    def _get_sequence_dense_tensor_state(column, features):
        sequence_feature_layer = sfc.SequenceFeatures(column)
        sequence_input, sequence_length = sequence_feature_layer(features)
        return sequence_input, sequence_length

    print(example_batch)
    sequence_input, sequence_length = _get_sequence_dense_tensor_state(thal_one_hot, example_batch)

    sequence_input = tf.keras.backend.sum(sequence_input, axis=-1)
    print(sequence_input)

    from notekeras.layers.embedding import EmbeddingRet
    res = EmbeddingRet(input_dim=6,
                       output_dim=11,
                       mask_zero=True,
                       name='Token-Embedding')(sequence_input)

    print(res)


example()
# SequenceEmbeddingColumnTes().run()
# SequenceSharedEmbeddingColumnTes2().run()
#
#
# SequenceCategoricalColumnWithIdentityTes().run()
# SequenceCategoricalColumnWithHashBucketTes().run()
# SequenceCategoricalColumnWithVocabularyListTes().run()



import tensorflow as tf

from notekeras.component import Component


class SPPNet(Component):
    def __init__(self, pool_size, strides=1, padding='SAME', data_format=None, *args, **kwargs):
        self.pool_size = pool_size
        self.strides = strides
        self.padding = padding
        self.data_format = data_format

        self.pool_list = None
        super(SPPNet, self).__init__(*args, **kwargs)

    def build(self, input_shape):
        self.pool_list = []

    def call(self, inputs, **kwargs):
        pool_list = []
        for ksize in self.pool_size:
            if ksize == 1:
                pool_list.append(inputs)
            else:
                pool_list.append(tf.nn.max_pool(inputs, ksize=ksize, padding=self.padding, strides=self.strides))

        output = tf.concat(pool_list, axis=-1)
        return output

    def compute_output_shape(self, input_shape):
        output_shape = input_shape
        output_shape[-1] = output_shape[-1] * len(self.pool_size)
        return output_shape



import pickle

import tensorflow as tf
from notekeras.layers import TrigPosEmbedding
from tensorflow.keras import backend as K
from tensorflow.keras.layers import (Concatenate, DenseFeatures, Embedding,
                                     Input, Layer)
from tensorflow.python.feature_column import feature_column_v2 as fc
from tensorflow.python.feature_column import sequence_feature_column as sfc
from tensorflow.python.feature_column.feature_column_lib import (
    NumericColumn, SequenceFeatures, numeric_column)

from .feature_column_def import IndicatorColumnDef

field_type_map = {
    'int': tf.dtypes.int32,
    'int32': tf.dtypes.int32,
    'int64': tf.dtypes.int64,
    'float': tf.dtypes.float32,
    'float32': tf.dtypes.float32,
    'float64': tf.dtypes.float64,
    'string': tf.dtypes.string,
}


def _parse_vocabulary(vocabulary):
    if isinstance(vocabulary, list):
        return vocabulary
    elif isinstance(vocabulary, str):
        return eval(vocabulary)
    else:
        return [i for i in vocabulary]


def _get_categorical_column(params: dict) -> fc.CategoricalColumn:
    if 'vocabulary' in params.keys():
        feature = fc.categorical_column_with_vocabulary_list(params['key'],
                                                             vocabulary_list=_parse_vocabulary(
                                                                 params['vocabulary']),
                                                             default_value=0)
    elif 'bucket_size' in params.keys():
        feature = fc.categorical_column_with_hash_bucket(params['key'],
                                                         hash_bucket_size=params['bucket_size'])
    elif 'file' in params.keys():
        feature = fc.categorical_column_with_vocabulary_file(params['key'],
                                                             vocabulary_file=params['file'],
                                                             default_value=0)
    elif 'num_buckets' in params.keys():
        feature = fc.categorical_column_with_identity(params['key'],
                                                      num_buckets=params['num_buckets'])
    elif 'boundaries' in params.keys():
        feature = fc.bucketized_column(fc.numeric_column(
            params['key']), boundaries=params['boundaries'])
    else:
        raise Exception("params error")

    return feature


def _get_sequence_categorical_column(params: dict) -> fc.SequenceCategoricalColumn:
    key = params['key']
    if 'vocabulary' in params.keys():
        feature = sfc.sequence_categorical_column_with_vocabulary_list(key,
                                                                       vocabulary_list=_parse_vocabulary(
                                                                           params['vocabulary']),
                                                                       default_value=0)
    elif 'bucket_size' in params.keys():
        feature = sfc.sequence_categorical_column_with_hash_bucket(
            key, hash_bucket_size=params['bucket_size'])
    elif 'file' in params.keys():
        feature = sfc.sequence_categorical_column_with_vocabulary_file(key,
                                                                       vocabulary_file=params['file'],
                                                                       default_value=0)
    elif 'num_buckets' in params.keys():
        feature = sfc.sequence_categorical_column_with_identity(key,
                                                                num_buckets=params['num_buckets'])
    else:
        raise Exception("params error")

    return feature


class ParseFeatureConfig:
    def __init__(self):
        self.feature_dict = {}
        self.share_layer = {}

    def _get_input_layer(self, params: dict, size=1) -> (str, Layer):
        field_name = params['key']
        if 'length' in params.keys():
            size = params['length']
        field_type = field_type_map.get(params['dtype'], tf.dtypes.string)

        if field_name in self.feature_dict.keys():
            inputs = self.feature_dict[field_name]
        else:
            inputs = Input((size,), dtype=field_type, name=field_name)
            self.feature_dict[field_name] = inputs

        return field_name, inputs

    def _get_share_layer(self, name: str, layer: Layer) -> Layer:
        """
        根据名称取出共享层
        :param name:
        :param layer:
        :return:
        """
        if name is None:
            return layer
        elif name in self.share_layer.keys():
            return self.share_layer[name]
        else:
            self.share_layer[name] = layer
            return layer

    def _numeric_column(self, params: dict) -> Layer:
        """
        输入：数值
        输出：数值
        :param params:
        :return:
        """
        key, inputs = self._get_input_layer(params)

        if 'transform' in params.keys():
            if params['transform'] == 'log':
                inputs = K.log(inputs)
            elif params['transform'] == 'sqrt':
                inputs = K.sqrt(inputs)

        #column = numeric_column(params['key'], shape=(params.get('length', 1),))
        #outputs = DenseFeatures(column)({key: inputs})
        return inputs

    def _cate_indicator_column(self, params: dict) -> DenseFeatures:
        """
        输入：类别
        输出：类别对应的one_hot
        :param params:
        :return:
        """
        key, inputs = self._get_input_layer(params)

        feature = _get_categorical_column(params)
        feature_column = fc.indicator_column(feature)

        outputs = DenseFeatures(
            feature_column, name=params.get('name', None))({key: inputs})

        return outputs

    def _cate_embedding_column(self, params: dict) -> Layer:
        """
        输入：类别
        输出：类别对应的embedding
        :param params:
        :return:
        """
        key, inputs = self._get_input_layer(params)

        feature = _get_categorical_column(params)

        column = IndicatorColumnDef(feature, size=1)

        sequence_input = DenseFeatures(column)({key: inputs})
        sequence_input = K.sum(sequence_input, axis=-1)

        name = params.get('share_name', None)
        layer = self._get_share_layer(name,
                                      Embedding(input_dim=feature.num_buckets + 1,
                                                output_dim=params['dimension'],
                                                mask_zero=True,
                                                # embeddings_regularizer=tf.keras.regularizers.l2(0.01),
                                                # activity_regularizer=tf.keras.regularizers.l2(0.01),
                                                name=name))
        res = layer(sequence_input)
        return res

    def _sequence_cate_indicator_column(self, params: dict):
        """
        输入：类别序列
        输出：类别序列对应的类别ID
        :param params:
        :return:
        """
        key, inputs = self._get_input_layer(params, size=params['length'])

        feature = _get_sequence_categorical_column(params)
        column = IndicatorColumnDef(feature, size=params['length'])

        sequence_input, sequence_length = SequenceFeatures(column)({
            key: inputs})

        return sequence_input, sequence_length

    def _sequence_cate_embedding_column(self, params: dict):
        """
        输入：类别序列
        输出：类别序列对应的embedding
        :param params:
        :return:
        """
        key, inputs = self._get_input_layer(params, size=params['length'])

        feature = _get_sequence_categorical_column(params)
        column = IndicatorColumnDef(feature, size=params['length'])

        sequence_input, sequence_length = SequenceFeatures(column)({
            key: inputs})

        sequence_input = tf.keras.backend.sum(sequence_input, axis=-1)

        name = params.get('share_name', None)
        layer = self._get_share_layer(name,
                                      Embedding(input_dim=feature.num_buckets + 1,
                                                output_dim=params['dimension'],
                                                mask_zero=True,
                                                weights=[pickle.load(open(params['weights'],
                                                                          'rb'))] if 'weights' in params.keys() else None,

                                                trainable=params.get(
                                                    'trainable', True),
                                                # embeddings_regularizer=tf.keras.regularizers.l2(0.01),
                                                # activity_regularizer=tf.keras.regularizers.l2(0.01),
                                                name=name))

        res = layer(sequence_input)
        res = TrigPosEmbedding(mode=TrigPosEmbedding.MODE_ADD, )(res)
        return res, sequence_length

    def _get_columns_map(self, key: str):
        _columns_map = {
            "NumericColumn": self._numeric_column,  # 数值类型
            "BucketizedColumn": self._cate_indicator_column,  # 分桶类型

            "CateIndicatorColumn": self._cate_indicator_column,  # 类别库生成的类别对应的one_hot
            "FileIndicatorColumn": self._cate_indicator_column,  # 读取文件产生的类别对应的one_hot
            "HashIndicatorColumn": self._cate_indicator_column,  # 关键词hash映射产生的类别对应的one_hot
            "BucketIndicatorColumn": self._cate_indicator_column,  # 数值分桶产生的类别对应的one_hot

            "CateEmbeddingColumn": self._cate_embedding_column,  # embedding
            "FileEmbeddingColumn": self._cate_embedding_column,
            "HashEmbeddingColumn": self._cate_embedding_column,
            "BucketEmbeddingColumn": self._cate_embedding_column,

            "SequenceCateIndicatorColumn": self._sequence_cate_indicator_column,
            "SequenceFileIndicatorColumn": self._sequence_cate_indicator_column,
            "SequenceHashIndicatorColumn": self._sequence_cate_indicator_column,

            "SequenceCateEmbddingColumn": self._sequence_cate_embedding_column,
            "SequenceFileEmbddingColumn": self._sequence_cate_embedding_column,
            "SequenceHashEmbddingColumn": self._sequence_cate_embedding_column
        }
        return _columns_map.get(key, None)

    def parse_feature_json(self, layer_json) -> Layer:
        outputs = []
        for feature_line in layer_json["inputs"]:
            feature_type_name = feature_line['type']
            feature_para = feature_line['parameters']

            method = self._get_columns_map(feature_type_name)
            if method is None:
                continue

            outputs.append(method(feature_para))

        #outputs = tf.keras.backend.concatenate(outputs)
        outputs = Concatenate()(outputs)

        return outputs

    def parse_sequence_feature_json(self, layer_json):
        feature_line = layer_json['inputs'][0]

        feature_type_name = feature_line['type']
        feature_para = feature_line['parameters']

        method = self._get_columns_map(feature_type_name)
        if method is None or not isinstance(feature_para, dict):
            raise Exception("error")

        sequence_input, sequence_length = method(feature_para)
        return sequence_input, sequence_length

    def parse_feature(self, layer_dict):
        values = []
        if isinstance(layer_dict, dict):
            values = layer_dict.values()
        elif isinstance(layer_dict, list):
            values = layer_dict

        outputs = []
        for value in values:
            if value['type'] == 'single':
                outputs.append(self.parse_feature_json(value))
            elif value['type'] == 'sequence':
                outputs.append(self.parse_sequence_feature_json(value))
        return outputs


def define_feature_json(key,
                        feature_type='NumericColumn',
                        feature_lenth=1,
                        dtype=None,
                        dimension=None,
                        share_name=None,
                        vocabulary=None,
                        bucket_size=None,
                        num_buckets=None,
                        *args, **kwargs):
    para = {
        "key": key,
        "length": feature_lenth,
        "dtype": dtype,
        "share_name": share_name,
        "dimension": dimension,
        "vocabulary": vocabulary,
        "bucket_size": bucket_size,
        "num_buckets": num_buckets
    }
    para.update(kwargs)
    para = {k: v for k, v in para.items() if k is not None and v is not None}

    return {
        "type": feature_type,
        "parameters": para
    }



import csv
import os
import os.path
import random
import sys
from collections import OrderedDict

import numpy as np
from PIL import Image
from six import raise_from
from tensorflow.keras import backend as K
from tensorflow.keras.utils import Sequence

from notekeras.models.retinanet.utils.anchors import anchor_targets_bbox, anchors_for_shape, guess_shapes
from notekeras.models.retinanet.utils.config import parse_anchor_parameters
from notekeras.models.retinanet.utils.image import TransformParameters, adjust_transform_for_image
from notekeras.models.retinanet.utils.image import apply_transform, preprocess_image, resize_image
from notekeras.models.retinanet.utils.image import read_image_bgr
from notekeras.models.retinanet.utils.transform import transform_aabb
from notekeras.utils import read_lines


class Generator(Sequence):

    def __init__(self,
                 transform_generator=None,
                 visual_effect_generator=None,
                 batch_size=1,
                 group_method='ratio',  # one of 'none', 'random', 'ratio'
                 shuffle_groups=True,
                 image_min_side=800,
                 image_max_side=1333,
                 no_resize=False,
                 transform_parameters=None,
                 compute_anchor_targets=anchor_targets_bbox,
                 compute_shapes=guess_shapes,
                 preprocess_image=preprocess_image,
                 config=None
                 ):
        """ Initialize Generator object.

        Args
            transform_generator    : A generator used to randomly transform images and annotations.
            batch_size             : The size of the batches to generate.
            group_method           : Determines how images are grouped together (defaults to 'ratio', one of ('none', 'random', 'ratio')).
            shuffle_groups         : If True, shuffles the groups each epoch.
            image_min_side         : After resizing the minimum side of an image is equal to image_min_side.
            image_max_side         : If after resizing the maximum side is larger than image_max_side, scales down further so that the max side is equal to image_max_side.
            no_resize              : If True, no image/annotation resizing is performed.
            transform_parameters   : The transform parameters used for data augmentation.
            compute_anchor_targets : Function handler for computing the targets of anchors for an image and its annotations.
            compute_shapes         : Function handler for computing the shapes of the pyramid for a given input.
            preprocess_image       : Function handler for preprocessing an image (scaling / normalizing) for passing through a network.
        """
        self.transform_generator = transform_generator
        self.visual_effect_generator = visual_effect_generator
        self.batch_size = int(batch_size)
        self.group_method = group_method
        self.shuffle_groups = shuffle_groups
        self.image_min_side = image_min_side
        self.image_max_side = image_max_side
        self.no_resize = no_resize
        self.transform_parameters = transform_parameters or TransformParameters()
        self.compute_anchor_targets = compute_anchor_targets
        self.compute_shapes = compute_shapes
        self.preprocess_image = preprocess_image
        self.config = config

        # Define groups
        self.group_images()

        # Shuffle when initializing
        if self.shuffle_groups:
            self.on_epoch_end()

    def on_epoch_end(self):
        if self.shuffle_groups:
            random.shuffle(self.groups)

    def size(self):
        """ Size of the dataset.
        """
        raise NotImplementedError('size method not implemented')

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        raise NotImplementedError('num_classes method not implemented')

    def has_label(self, label):
        """ Returns True if label is a known label.
        """
        raise NotImplementedError('has_label method not implemented')

    def has_name(self, name):
        """ Returns True if name is a known class.
        """
        raise NotImplementedError('has_name method not implemented')

    def name_to_label(self, name):
        """ Map name to label.
        """
        raise NotImplementedError('name_to_label method not implemented')

    def label_to_name(self, label):
        """ Map label to name.
        """
        raise NotImplementedError('label_to_name method not implemented')

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        raise NotImplementedError('image_aspect_ratio method not implemented')

    def image_path(self, image_index):
        """ Get the path to an image.
        """
        raise NotImplementedError('image_path method not implemented')

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        raise NotImplementedError('load_image method not implemented')

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        raise NotImplementedError('load_annotations method not implemented')

    def load_annotations_group(self, group):
        """ Load annotations for all images in group.
        """
        annotations_group = [self.load_annotations(image_index) for image_index in group]
        for annotations in annotations_group:
            assert (isinstance(annotations,
                               dict)), '\'load_annotations\' should return a list of dictionaries, received: {}'.format(
                type(annotations))
            assert (
                    'labels' in annotations), '\'load_annotations\' should return a list of dictionaries that contain \'labels\' and \'bboxes\'.'
            assert (
                    'bboxes' in annotations), '\'load_annotations\' should return a list of dictionaries that contain \'labels\' and \'bboxes\'.'

        return annotations_group

    def filter_annotations(self, image_group, annotations_group, group):
        """ Filter annotations by removing those that are outside of the image bounds or whose width/height < 0.
        """
        # test all annotations
        for index, (image, annotations) in enumerate(zip(image_group, annotations_group)):
            # test x2 < x1 | y2 < y1 | x1 < 0 | y1 < 0 | x2 <= 0 | y2 <= 0 | x2 >= image.shape[1] | y2 >= image.shape[0]
            invalid_indices = np.where(
                (annotations['bboxes'][:, 2] <= annotations['bboxes'][:, 0]) |
                (annotations['bboxes'][:, 3] <= annotations['bboxes'][:, 1]) |
                (annotations['bboxes'][:, 0] < 0) |
                (annotations['bboxes'][:, 1] < 0) |
                (annotations['bboxes'][:, 2] > image.shape[1]) |
                (annotations['bboxes'][:, 3] > image.shape[0]))[0]

            # delete invalid indices
            if len(invalid_indices):
                # warnings.warn('Image {} with id {} (shape {}) contains the following invalid boxes: {}.'.format(
                #     self.image_path(group[index]),
                #     group[index],
                #     image.shape,
                #     annotations['bboxes'][invalid_indices, :]
                # ))
                for k in annotations_group[index].keys():
                    annotations_group[index][k] = np.delete(annotations[k], invalid_indices, axis=0)
        return image_group, annotations_group

    def load_image_group(self, group):
        """ Load images for all images in a group.
        """
        return [self.load_image(image_index) for image_index in group]

    def random_visual_effect_group_entry(self, image, annotations):
        """ Randomly transforms image and annotation.
        """
        visual_effect = next(self.visual_effect_generator)
        # apply visual effect
        image = visual_effect(image)
        return image, annotations

    def random_visual_effect_group(self, image_group, annotations_group):
        """ Randomly apply visual effect on each image.
        """
        assert (len(image_group) == len(annotations_group))

        if self.visual_effect_generator is None:
            # do nothing
            return image_group, annotations_group

        for index in range(len(image_group)):
            # apply effect on a single group entry
            image_group[index], annotations_group[index] = self.random_visual_effect_group_entry(
                image_group[index], annotations_group[index]
            )

        return image_group, annotations_group

    def random_transform_group_entry(self, image, annotations, transform=None):
        """ Randomly transforms image and annotation.
        """
        # randomly transform both image and annotations
        if transform is not None or self.transform_generator:
            if transform is None:
                transform = adjust_transform_for_image(next(self.transform_generator), image,
                                                       self.transform_parameters.relative_translation)

            # apply transformation to image
            image = apply_transform(transform, image, self.transform_parameters)

            # Transform the bounding boxes in the annotations.
            annotations['bboxes'] = annotations['bboxes'].copy()
            for index in range(annotations['bboxes'].shape[0]):
                annotations['bboxes'][index, :] = transform_aabb(transform, annotations['bboxes'][index, :])

        return image, annotations

    def random_transform_group(self, image_group, annotations_group):
        """ Randomly transforms each image and its annotations.
        """

        assert (len(image_group) == len(annotations_group))

        for index in range(len(image_group)):
            # transform a single group entry
            image_group[index], annotations_group[index] = self.random_transform_group_entry(image_group[index],
                                                                                             annotations_group[index])

        return image_group, annotations_group

    def resize_image(self, image):
        """ Resize an image using image_min_side and image_max_side.
        """
        if self.no_resize:
            return image, 1
        else:
            return resize_image(image, min_side=self.image_min_side, max_side=self.image_max_side)

    def preprocess_group_entry(self, image, annotations):
        """ Preprocess image and its annotations.
        """
        # preprocess the image
        image = self.preprocess_image(image)

        # resize image
        image, image_scale = self.resize_image(image)

        # apply resizing to annotations too
        annotations['bboxes'] *= image_scale

        # convert to the wanted keras floatx
        image = K.cast_to_floatx(image)

        return image, annotations

    def preprocess_group(self, image_group, annotations_group):
        """ Preprocess each image and its annotations in its group.
        """
        assert (len(image_group) == len(annotations_group))

        for index in range(len(image_group)):
            # preprocess a single group entry
            image_group[index], annotations_group[index] = self.preprocess_group_entry(image_group[index],
                                                                                       annotations_group[index])

        return image_group, annotations_group

    def group_images(self):
        """ Order the images according to self.order and makes groups of self.batch_size.
        """
        # determine the order of the images
        order = list(range(self.size()))
        if self.group_method == 'random':
            random.shuffle(order)
        elif self.group_method == 'ratio':
            order.sort(key=lambda x: self.image_aspect_ratio(x))

        # divide into groups, one group = one batch
        self.groups = [[order[x % len(order)] for x in range(i, i + self.batch_size)] for i in
                       range(0, len(order), self.batch_size)]

    def compute_inputs(self, image_group):
        """ Compute inputs for the network using an image_group.
        """
        # get the max image shape
        max_shape = tuple(max(image.shape[x] for image in image_group) for x in range(3))

        # construct an image batch object
        image_batch = np.zeros((self.batch_size,) + max_shape, dtype=K.floatx())

        # copy all images to the upper left part of the image batch object
        for image_index, image in enumerate(image_group):
            image_batch[image_index, :image.shape[0], :image.shape[1], :image.shape[2]] = image

        if K.image_data_format() == 'channels_first':
            image_batch = image_batch.transpose((0, 3, 1, 2))

        return image_batch

    def generate_anchors(self, image_shape):
        anchor_params = None
        if self.config and 'anchor_parameters' in self.config:
            anchor_params = parse_anchor_parameters(self.config)
        return anchors_for_shape(image_shape, anchor_params=anchor_params, shapes_callback=self.compute_shapes)

    def compute_targets(self, image_group, annotations_group):
        """ Compute target outputs for the network using images and their annotations.
        """
        # get the max image shape
        max_shape = tuple(max(image.shape[x] for image in image_group) for x in range(3))
        anchors = self.generate_anchors(max_shape)

        batches = self.compute_anchor_targets(anchors, image_group, annotations_group, self.num_classes())

        return list(batches)

    def compute_input_output(self, group):
        """ Compute inputs and target outputs for the network.
        """
        # load images and annotations
        image_group = self.load_image_group(group)
        annotations_group = self.load_annotations_group(group)

        # check validity of annotations
        image_group, annotations_group = self.filter_annotations(image_group, annotations_group, group)

        # randomly apply visual effect
        image_group, annotations_group = self.random_visual_effect_group(image_group, annotations_group)

        # randomly transform data
        image_group, annotations_group = self.random_transform_group(image_group, annotations_group)

        # perform preprocessing steps
        image_group, annotations_group = self.preprocess_group(image_group, annotations_group)

        # compute network inputs
        inputs = self.compute_inputs(image_group)

        # compute network targets
        targets = self.compute_targets(image_group, annotations_group)

        return inputs, targets

    def __len__(self):
        """
        Number of batches for generator.
        """

        return len(self.groups)

    def __getitem__(self, index):
        group = self.groups[index]
        inputs, targets = self.compute_input_output(group)

        return inputs, targets


class CSVGenerator(Generator):
    """ Generate data for a custom CSV dataset.

    See https://github.com/fizyr/keras-retinanet#csv-datasets for more information.
    """

    def __init__(self, csv_data_file, csv_class_file, base_dir=None, **kwargs):
        """ Initialize a CSV data generator.

        Args
            csv_data_file: Path to the CSV annotations file.
            csv_class_file: Path to the CSV classes file.
            base_dir: Directory w.r.t. where the files are to be searched (defaults to the directory containing the csv_data_file).
        """
        self.image_names = []
        self.image_data = {}
        self.base_dir = base_dir
        self.csv_data_file = csv_data_file

        # Take base_dir from annotations file if not explicitly specified.
        if self.base_dir is None:
            self.base_dir = os.path.dirname(csv_data_file)

        # parse the provided class file
        try:
            with self._open_for_csv(csv_class_file) as file:
                self.classes = self._read_classes(csv.reader(file, delimiter=','))
        except ValueError as e:
            raise_from(ValueError('invalid CSV class file: {}: {}'.format(csv_class_file, e)), None)

        self.labels = {}
        for key, value in self.classes.items():
            self.labels[value] = key

        # csv with img_path, x1, y1, x2, y2, class_name
        try:
            with self._open_for_csv(csv_data_file) as file:
                self.image_data = self._read_annotations(csv.reader(file, delimiter=','), self.classes)
        except ValueError as e:
            raise_from(ValueError('invalid CSV annotations file: {}: {}'.format(csv_data_file, e)), None)
        self.image_names = list(self.image_data.keys())

        super(CSVGenerator, self).__init__(**kwargs)

    def size(self):
        """ Size of the dataset.
        """
        return len(self.image_names)

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        return max(self.classes.values()) + 1

    def has_label(self, label):
        """ Return True if label is a known label.
        """
        return label in self.labels

    def has_name(self, name):
        """ Returns True if name is a known class.
        """
        return name in self.classes

    def name_to_label(self, name):
        """ Map name to label.
        """
        return self.classes[name]

    def label_to_name(self, label):
        """ Map label to name.
        """
        return self.labels[label]

    def image_path(self, image_index):
        """ Returns the image path for image_index.
        """
        return os.path.join(self.base_dir, self.image_names[image_index])

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        # PIL is fast for metadata
        image = Image.open(self.image_path(image_index))
        return float(image.width) / float(image.height)

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        return read_image_bgr(self.image_path(image_index))

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        path = self.image_names[image_index]
        annotations = {'labels': np.empty((0,)), 'bboxes': np.empty((0, 4))}

        for idx, annot in enumerate(self.image_data[path]):
            annotations['labels'] = np.concatenate((annotations['labels'], [self.name_to_label(annot['class'])]))
            annotations['bboxes'] = np.concatenate((annotations['bboxes'], [[
                float(annot['x1']),
                float(annot['y1']),
                float(annot['x2']),
                float(annot['y2']),
            ]]))

        return annotations

    @staticmethod
    def _open_for_csv(path):
        """ Open a file with flags suitable for csv.reader.

        This is different for python2 it means with mode 'rb',
        for python3 this means 'r' with "universal newlines".
        """
        if sys.version_info[0] < 3:
            return open(path, 'rb')
        else:
            return open(path, 'r', newline='')

    def _read_annotations(self, csv_reader, classes):
        result = OrderedDict()
        for line, row in enumerate(csv_reader):
            line += 1

            try:
                img_file, x1, y1, x2, y2, class_name = row[:6]
            except ValueError:
                raise_from(ValueError(
                    "line {}: format should be 'img_file,x1,y1,x2,y2,class_name' or 'img_file,,,,,'".format(line)),
                    None)

            if img_file not in result:
                result[img_file] = []

            # If a row contains only an image path, it's an image without annotations.
            if (x1, y1, x2, y2, class_name) == ('', '', '', '', ''):
                continue

            x1 = self._parse(x1, int, 'line {}: malformed x1: {{}}'.format(line))
            y1 = self._parse(y1, int, 'line {}: malformed y1: {{}}'.format(line))
            x2 = self._parse(x2, int, 'line {}: malformed x2: {{}}'.format(line))
            y2 = self._parse(y2, int, 'line {}: malformed y2: {{}}'.format(line))

            if x2 <= x1:
                raise ValueError('line {}: x2 ({}) must be higher than x1 ({})'.format(line, x2, x1))
            if y2 <= y1:
                raise ValueError('line {}: y2 ({}) must be higher than y1 ({})'.format(line, y2, y1))

            if class_name not in classes:
                raise ValueError('line {}: unknown class name: \'{}\' (classes: {})'.format(line, class_name, classes))

            result[img_file].append({'x1': x1, 'x2': x2, 'y1': y1, 'y2': y2, 'class': class_name})
        return result

    @staticmethod
    def _parse(value, function, fmt):
        """
        Parse a string into a value, and format a nice ValueError if it fails.

        Returns `function(value)`.
        Any `ValueError` raised is catched and a new `ValueError` is raised
        with message `fmt.format(e)`, where `e` is the caught `ValueError`.
        """
        try:
            return function(value)
        except ValueError as e:
            raise_from(ValueError(fmt.format(e)), None)

    def _read_classes(self, csv_reader):
        """ Parse the classes file given by csv_reader.
        """
        result = OrderedDict()
        for line, row in enumerate(csv_reader):
            line += 1

            try:
                class_name, class_id = row
            except ValueError:
                raise_from(ValueError('line {}: format should be \'class_name,class_id\''.format(line)), None)
            class_id = self._parse(class_id, int, 'line {}: malformed class ID: {{}}'.format(line))

            if class_name in result:
                raise ValueError('line {}: duplicate class name: \'{}\''.format(line, class_name))
            result[class_name] = class_id
        return result


class TextGenerator(Generator):
    """ Generate data for a custom CSV dataset.

    See https://github.com/fizyr/keras-retinanet#csv-datasets for more information.
    """

    def __init__(self, csv_data_file, csv_class_file, base_dir=None, **kwargs):
        """ Initialize a CSV data generator.

        Args
            csv_data_file: Path to the CSV annotations file.
            csv_class_file: Path to the CSV classes file.
            base_dir: Directory w.r.t. where the files are to be searched (defaults to the directory containing the csv_data_file).
        """
        self.class_names = []
        self.image_names = []
        self.image_data = {}
        self.base_dir = base_dir
        self.csv_data_file = csv_data_file
        self.csv_class_file = csv_class_file

        # Take base_dir from annotations file if not explicitly specified.
        if self.base_dir is None:
            self.base_dir = os.path.dirname(csv_data_file)

        # parse the provided class file
        try:
            self.class_names = read_lines(csv_class_file)
            self.classes = {}
            for i, name in enumerate(self.class_names):
                self.classes[name] = i
        except ValueError as e:
            raise_from(ValueError('invalid CSV class file: {}: {}'.format(csv_class_file, e)), None)

        self.labels = {}
        for key, value in self.classes.items():
            self.labels[value] = key

        # csv with img_path, x1, y1, x2, y2, class_name
        try:
            self.image_data = self._read_annotations()
        except ValueError as e:
            raise_from(ValueError('invalid CSV annotations file: {}: {}'.format(csv_data_file, e)), None)
        self.image_names = list(self.image_data.keys())

        super(TextGenerator, self).__init__(**kwargs)

    def size(self):
        """ Size of the dataset.
        """
        return len(self.image_names)

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        return max(self.classes.values()) + 1

    def has_label(self, label):
        """ Return True if label is a known label.
        """
        return label in self.labels

    def has_name(self, name):
        """ Returns True if name is a known class.
        """
        return name in self.classes

    def name_to_label(self, name):
        """ Map name to label.
        """
        return self.classes[name]

    def label_to_name(self, label):
        """ Map label to name.
        """
        return self.labels[label]

    def image_path(self, image_index):
        """ Returns the image path for image_index.
        """
        return os.path.join(self.base_dir, self.image_names[image_index])

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        # PIL is fast for metadata
        image = Image.open(self.image_path(image_index))
        return float(image.width) / float(image.height)

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        return read_image_bgr(self.image_path(image_index))

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        path = self.image_names[image_index]
        annotations = {'labels': np.empty((0,)), 'bboxes': np.empty((0, 4))}

        for idx, annot in enumerate(self.image_data[path]):
            annotations['labels'] = np.concatenate((annotations['labels'], [self.name_to_label(annot['class'])]))
            annotations['bboxes'] = np.concatenate((annotations['bboxes'], [[
                float(annot['x1']),
                float(annot['y1']),
                float(annot['x2']),
                float(annot['y2']), ]]))

        return annotations

    @staticmethod
    def _open_for_csv(path):
        """ Open a file with flags suitable for csv.reader.

        This is different for python2 it means with mode 'rb',
        for python3 this means 'r' with "universal newlines".
        """
        if sys.version_info[0] < 3:
            return open(path, 'rb')
        else:
            return open(path, 'r', newline='')

    def _read_annotations(self):
        result = OrderedDict()
        for annotation in read_lines(self.csv_data_file):
            rows = annotation.split(' ')
            img_file = rows[0]
            if img_file not in result:
                result[img_file] = []

            for box in rows[1:]:
                if len(box) < 5:
                    continue
                x1, y1, x2, y2, class_name = box.split(',')
                x1, y1, x2, y2 = int(x1), int(y1), int(x2), int(y2)
                class_name = self.class_names[int(class_name)]
                result[img_file].append({'x1': x1, 'x2': x2, 'y1': y1, 'y2': y2, 'class': class_name})
        print('read annotations done')
        return result

    @staticmethod
    def _parse(value, function, fmt):
        try:
            return function(value)
        except ValueError as e:
            raise_from(ValueError(fmt.format(e)), None)


class KittiGenerator(Generator):
    kitti_classes = {
        'Car': 0,
        'Van': 1,
        'Truck': 2,
        'Pedestrian': 3,
        'Person_sitting': 4,
        'Cyclist': 5,
        'Tram': 6,
        'Misc': 7,
        'DontCare': 7
    }
    """ Generate data for a KITTI dataset.

    See http://www.cvlibs.net/datasets/kitti/ for more information.
    """

    def __init__(
            self,
            base_dir,
            subset='train',
            **kwargs
    ):
        """ Initialize a KITTI data generator.

        Args
            base_dir: Directory w.r.t. where the files are to be searched (defaults to the directory containing the csv_data_file).
            subset: The subset to generate data for (defaults to 'train').
        """
        self.base_dir = base_dir

        label_dir = os.path.join(self.base_dir, subset, 'labels')
        image_dir = os.path.join(self.base_dir, subset, 'images')

        """
        1    type         Describes the type of object: 'Car', 'Van', 'Truck',
                             'Pedestrian', 'Person_sitting', 'Cyclist', 'Tram',
                             'Misc' or 'DontCare'
        1    truncated    Float from 0 (non-truncated) to 1 (truncated), where
                         truncated refers to the object leaving image boundaries
        1    occluded     Integer (0,1,2,3) indicating occlusion state:
                         0 = fully visible, 1 = partly occluded
                         2 = largely occluded, 3 = unknown
        1    alpha        Observation angle of object, ranging [-pi..pi]
        4    bbox         2D bounding box of object in the image (0-based index):
                         contains left, top, right, bottom pixel coordinates
        3    dimensions   3D object dimensions: height, width, length (in meters)
        3    location     3D object location x,y,z in camera coordinates (in meters)
        1    rotation_y   Rotation ry around Y-axis in camera coordinates [-pi..pi]
        """

        self.labels = {}
        self.classes = self.kitti_classes
        for name, label in self.classes.items():
            self.labels[label] = name

        self.image_data = dict()
        self.images = []
        for i, fn in enumerate(os.listdir(label_dir)):
            label_fp = os.path.join(label_dir, fn)
            image_fp = os.path.join(image_dir, fn.replace('.txt', '.png'))

            self.images.append(image_fp)

            fieldnames = ['type', 'truncated', 'occluded', 'alpha', 'left', 'top', 'right', 'bottom', 'dh', 'dw', 'dl',
                          'lx', 'ly', 'lz', 'ry']
            with open(label_fp, 'r') as csv_file:
                reader = csv.DictReader(csv_file, delimiter=' ', fieldnames=fieldnames)
                boxes = []
                for line, row in enumerate(reader):
                    label = row['type']
                    cls_id = self.kitti_classes[label]

                    annotation = {'cls_id': cls_id, 'x1': row['left'], 'x2': row['right'], 'y2': row['bottom'],
                                  'y1': row['top']}
                    boxes.append(annotation)

                self.image_data[i] = boxes

        super(KittiGenerator, self).__init__(**kwargs)

    def size(self):
        """ Size of the dataset.
        """
        return len(self.images)

    def num_classes(self):
        """ Number of classes in the dataset.
        """
        return max(self.classes.values()) + 1

    def has_label(self, label):
        """ Return True if label is a known label.
        """
        return label in self.labels

    def has_name(self, name):
        """ Returns True if name is a known class.
        """
        return name in self.classes

    def name_to_label(self, name):
        """ Map name to label.
        """
        raise NotImplementedError()

    def label_to_name(self, label):
        """ Map label to name.
        """
        return self.labels[label]

    def image_aspect_ratio(self, image_index):
        """ Compute the aspect ratio for an image with image_index.
        """
        # PIL is fast for metadata
        image = Image.open(self.images[image_index])
        return float(image.width) / float(image.height)

    def image_path(self, image_index):
        """ Get the path to an image.
        """
        return self.images[image_index]

    def load_image(self, image_index):
        """ Load an image at the image_index.
        """
        return read_image_bgr(self.image_path(image_index))

    def load_annotations(self, image_index):
        """ Load annotations for an image_index.
        """
        image_data = self.image_data[image_index]
        annotations = {'labels': np.empty((len(image_data),)), 'bboxes': np.empty((len(image_data), 4))}

        for idx, ann in enumerate(image_data):
            annotations['bboxes'][idx, 0] = float(ann['x1'])
            annotations['bboxes'][idx, 1] = float(ann['y1'])
            annotations['bboxes'][idx, 2] = float(ann['x2'])
            annotations['bboxes'][idx, 3] = float(ann['y2'])
            annotations['labels'][idx] = int(ann['cls_id'])

        return annotations



import tensorflow as tf
from tensorflow.keras import backend as K


def focal(alpha=0.25, gamma=2.0):
    """ Create a functor for computing the focal loss.

    Args
        alpha: Scale the focal weight with alpha.
        gamma: Take the power of the focal weight with gamma.

    Returns
        A functor that computes the focal loss using the alpha and gamma.
    """

    def _focal(y_true, y_pred):
        """ Compute the focal loss given the target tensor and the predicted tensor.

        As defined in https://arxiv.org/abs/1708.02002

        Args
            y_true: Tensor of target data from the generator with shape (B, N, num_classes).
            y_pred: Tensor of predicted data from the network with shape (B, N, num_classes).

        Returns
            The focal loss of y_pred w.r.t. y_true.
        """
        labels = y_true[:, :, :-1]
        anchor_state = y_true[:, :, -1]  # -1 for ignore, 0 for background, 1 for object
        classification = y_pred

        # filter out "ignore" anchors
        indices = tf.where(K.not_equal(anchor_state, -1))
        labels = tf.gather_nd(labels, indices)
        classification = tf.gather_nd(classification, indices)

        # compute the focal loss
        alpha_factor = K.ones_like(labels) * alpha
        alpha_factor = tf.where(K.equal(labels, 1), alpha_factor, 1 - alpha_factor)
        focal_weight = tf.where(K.equal(labels, 1), 1 - classification, classification)
        focal_weight = alpha_factor * focal_weight ** gamma

        cls_loss = focal_weight * K.binary_crossentropy(labels, classification)

        # compute the normalizer: the number of positive anchors
        normalizer = tf.where(K.equal(anchor_state, 1))
        normalizer = K.cast(K.shape(normalizer)[0], K.floatx())
        normalizer = K.maximum(K.cast_to_floatx(1.0), normalizer)

        return K.sum(cls_loss) / normalizer

    return _focal


def smooth_l1(sigma=3.0):
    """ Create a smooth L1 loss functor.

    Args
        sigma: This argument defines the point where the loss changes from L2 to L1.

    Returns
        A functor for computing the smooth L1 loss given target data and predicted data.
    """
    sigma_squared = sigma ** 2

    def _smooth_l1(y_true, y_pred):
        """ Compute the smooth L1 loss of y_pred w.r.t. y_true.

        Args
            y_true: Tensor from the generator of shape (B, N, 5). The last value for each box is the state of the anchor (ignore, negative, positive).
            y_pred: Tensor from the network of shape (B, N, 4).

        Returns
            The smooth L1 loss of y_pred w.r.t. y_true.
        """
        # separate target and state
        regression = y_pred
        regression_target = y_true[:, :, :-1]
        anchor_state = y_true[:, :, -1]

        # filter out "ignore" anchors
        indices = tf.where(K.equal(anchor_state, 1))
        regression = tf.gather_nd(regression, indices)
        regression_target = tf.gather_nd(regression_target, indices)

        # compute smooth L1 loss
        # f(x) = 0.5 * (sigma * x)^2          if |x| < 1 / sigma / sigma
        #        |x| - 0.5 / sigma / sigma    otherwise
        regression_diff = regression - regression_target
        regression_diff = K.abs(regression_diff)
        regression_loss = tf.where(
            K.less(regression_diff, 1.0 / sigma_squared),
            0.5 * sigma_squared * K.pow(regression_diff, 2),
            regression_diff - 0.5 / sigma_squared
        )

        # compute the normalizer: the number of positive anchors
        normalizer = K.maximum(1, K.shape(indices)[0])
        normalizer = K.cast(normalizer, dtype=K.floatx())
        return K.sum(regression_loss) / normalizer

    return _smooth_l1



from collections import Counter

import numpy as np
import pandas as pd
import tensorflow as tf
from sklearn.model_selection import train_test_split

pd.set_option('display.max_columns', None)
pd.set_option('display.max_rows', 200)
pd.set_option('max_colwidth', 500)


def agg_set(size=-1, padding=False):
    def inner(x):
        if size <= 0:
            return list(set(x))
        elif padding is False:
            return list(set(x))[-size:]
        else:
            x = list(set(x))
            return list(x[-size:] + ['0'] * (size - len(x)))

    return inner


def agg_list(size=-1, padding=False):
    # def inner(x):
    #     if size <= 0:
    #         return list(x)
    #     elif padding is False:
    #         return list(x)[-size:]
    #     else:
    #         x = list(x)
    #         return list(x[-size:] + ['0'] * (size - len(x)))
    #
    # return inner

    if size <= 0:
        return lambda x: list(x)
    elif padding is False:
        return lambda x: list(x)[-size:]
    else:
        return lambda x: list(list(x)[-size:] + ['0'] * (size - len(list(x))))


print(agg_list(size=10, padding=True)([1, 2, 5]))


def agg_array(size=-1, pct=True):
    def inner(x):
        category_list = np.zeros(size)
        for category in list(x):
            category_list[int(category) - 1] += 1

        return category_list.tolist()

    def inner2(x):
        category_list = np.zeros(size)
        for category in list(x):
            category_list[int(category) - 1] += 1

        category_list = category_list / category_list.sum()
        category_list = np.round(category_list, 4)
        return category_list.tolist()

    if pct:
        return inner2
    else:
        return inner
    # def inner(x):
    #     if size <= 0:
    #         return 0
    #     category_list = np.zeros(size)
    #     for category in list(x):
    #         category_list[int(category) - 1] += 1
    #
    #     if pct:
    #         category_list = category_list / category_list.sum()
    #         category_list = np.round(category_list, 4)
    #
    #     return category_list.tolist()
    #
    # return inner


def agg_unique_mean():
    def inner(x):
        return np.mean(list(Counter(list(x)).values()))

    return inner


def agg_unique_var():
    def inner(x):
        return np.var(list(Counter(list(x)).values()))

    return inner


def agg_fun_def(agg_name: str, params: dict = None):
    agg_fun_dict = {
        'size': 'size',
        'count': 'count',
        'mean': 'mean',
        'unique': 'nunique',
        'nunique': 'nunique',
        'max': 'max',
        'min': 'min',
        'sum': 'sum',
        'std': 'std',
        'median': 'median',
        'skew': 'skew',
        'list': agg_list,
        'set': agg_set,
        'array': agg_array,
        'unique_mean': agg_unique_mean,
        'unique_var': agg_unique_var,
    }
    if agg_name not in agg_fun_dict:
        return 'size'

    agg_func = agg_fun_dict[agg_name]
    if isinstance(agg_func, str):
        return agg_func
    elif params is None:
        return agg_func()
    else:
        return agg_func(**params)


def config_agg(configs: dict):
    agg_info = {}
    for stat_col in configs['stat_cols'].keys():
        stat_values = configs['stat_cols'][stat_col]

        cols = []
        for stat_value in stat_values:
            col_name = stat_value['col_name']
            agg_func = stat_value['agg_func']
            agg_para = stat_value.get('agg_para', None)

            agg_func = agg_fun_def(agg_func, agg_para)

            cols.append((col_name, agg_func))
        agg_info[stat_col] = cols

    return agg_info


def run():
    path_root = "/Users/liangtaoniu/tmp/dataset/tencent2020/train_preliminary"
    path_ad = path_root + '/ad.csv'
    path_user = path_root + '/user.csv'
    path_click = path_root + '/click_log.csv'

    feature_pandas = {
        "feature1": {
            "group_key": ["user_id"],
            "stat_cols": {
                "time": [
                    {"agg_func": "unique", "col_name": "time_unique"},
                    {"agg_func": "count", "col_name": "time_count"},
                    {"agg_func": "list", "col_name": "time_list", "agg_para": {"size": 10, "padding": True}},
                    {"agg_func": "array", "col_name": "time_pct", "agg_para": {"size": 91, "pct": True}}
                ],
                "advertiser_id": [
                    {"agg_func": "unique", "col_name": "advertiser_unique"},
                    {"agg_func": "count", "col_name": "advertiser_count"},
                    {"agg_func": "list", "col_name": "advertiser_list", "agg_para": {"size": 10, "padding": True}},
                ],
                "creative_id": [
                    {"agg_func": "unique", "col_name": "creative_unique"},
                    {"agg_func": "count", "col_name": "creative_count"},
                    {"agg_func": "list", "col_name": "creative_list", "agg_para": {"size": 10, "padding": True}},
                ],

                "ad_id": [
                    {"agg_func": "unique", "col_name": "ad_unique"},
                    {"agg_func": "count", "col_name": "ad_count"},
                    {"agg_func": "list", "col_name": "ad_list", "agg_para": {"size": 10, "padding": True}},
                ],

                "product_id": [
                    {"agg_func": "unique", "col_name": "product_unique"},
                    {"agg_func": "count", "col_name": "product_count"},
                    {"agg_func": "list", "col_name": "product_list", "agg_para": {"size": 10, "padding": True}},
                ],

                "product_category": [
                    {"agg_func": "unique", "col_name": "category_unique"},
                    {"agg_func": "count", "col_name": "category_count"},
                    {"agg_func": "list", "col_name": "category_list", "agg_para": {"size": 10, "padding": True}},
                    {"agg_func": "array", "col_name": "category_pct", "agg_para": {"size": 18, "pct": True}}
                ],

                "industry": [
                    {"agg_func": "unique", "col_name": "industry_unique"},
                    {"agg_func": "count", "col_name": "industry_count"},
                    {"agg_func": "list", "col_name": "industry_list", "agg_para": {"size": 10, "padding": True}},
                    {"agg_func": "array", "col_name": "industry_pct", "agg_para": {"size": 335, "pct": True}}
                ],
                "weekday": [
                    {"agg_func": "unique", "col_name": "weekday_unique"},
                    {"agg_func": "count", "col_name": "weekday_count"},
                    {"agg_func": "list", "col_name": "weekday_list", "agg_para": {"size": 10, "padding": True}},
                    {"agg_func": "array", "col_name": "weekday_pct", "agg_para": {"size": 7, "pct": True}}
                ]
            }
        }
    }

    feature_pandas2 = {
        "feature1": {
            "group_key": ["user_id"],
            "stat_cols": {
                "creative_id": [
                    {"agg_func": "nunique", "col_name": "creative_unique"},
                    {"agg_func": "count", "col_name": "creative_count"},
                    {"agg_func": "list", "col_name": "creative_list", "agg_para": {"size": 10, "padding": True}}
                ],
                "click_times": [
                    {"agg_func": "nunique", "col_name": "click_unique"},
                    {"agg_func": "count", "col_name": "click_count"},
                    {"agg_func": "list", "col_name": "click_list", "agg_para": {"size": 10, "padding": True}}
                ],
                "time": [
                    {"agg_func": "nunique", "col_name": "time_unique"},
                    {"agg_func": "count", "col_name": "time_count"},
                    {"agg_func": "list", "col_name": "time_list", "agg_para": {"size": 10, "padding": True}},
                    {"agg_func": "array", "col_name": "time_pct", "agg_para": {"size": 99, "pct": True}}
                ],
            }
        }
    }

    df_ad = pd.read_csv(path_ad)
    df_user = pd.read_csv(path_user)
    df_click = pd.read_csv(path_click).head(200 * 100)

    df_ad.loc[df_ad['product_id'] == '\\N', 'product_id'] = 0
    df_ad.loc[df_ad['product_category'] == '\\N', 'product_category'] = 0
    df_ad.loc[df_ad['advertiser_id'] == '\\N', 'advertiser_id'] = 0
    df_ad.loc[df_ad['industry'] == '\\N', 'industry'] = 0

    df_click = df_click.sort_values(['user_id', 'time'])

    pd_merge = pd.merge(df_click, df_ad, on='creative_id')
    pd_merge = pd.merge(pd_merge, df_user, on='user_id')

    pd_merge['weekday'] = pd_merge['time'] % 7 + 1
    pd_merge['product_id'] = pd_merge['product_id'].astype(str)
    pd_merge['product_category'] = pd_merge['product_category'].astype(str)
    pd_merge['advertiser_id'] = pd_merge['advertiser_id'].astype(str)
    pd_merge['industry'] = pd_merge['industry'].astype(str)

    agg_info = config_agg(feature_pandas['feature1'])

    dataframe = pd_merge.groupby(['user_id']).agg(agg_info)
    dataframe = pd.merge(dataframe, df_user, on='user_id')

    def df_to_dataset(dataframe, shuffle=True, batch_size=32):
        dataframe = dataframe.copy()
        label1 = tf.keras.backend.one_hot(dataframe.pop('gender'), 2)
        label2 = tf.keras.backend.one_hot(dataframe.pop('age'), 10)

        labels = (label1, label2)
        ds = tf.data.Dataset.from_tensor_slices((dict(dataframe), labels))
        if shuffle:
            ds = ds.shuffle(buffer_size=len(dataframe))
        ds = ds.batch(batch_size)
        return ds

    batch_size = 256
    train, test = train_test_split(dataframe, test_size=0.2)
    train, val = train_test_split(train, test_size=0.2)

    train_d = df_to_dataset(train, batch_size=batch_size)
    val_d = df_to_dataset(val, shuffle=False, batch_size=batch_size)
    test_d = df_to_dataset(test, shuffle=False, batch_size=batch_size)



from tensorflow import keras
from tensorflow.keras.utils import get_file

from . import Backbone
from .retinanet import RetinaNetModel
from ..utils.image import preprocess_image


class VGGBackbone(Backbone):
    """ Describes backbone information and provides utility functions.
    """

    def retinanet(self, *args, **kwargs):
        """ Returns a retinanet model using the correct backbone.
        """
        return vgg_retinanet(*args, backbone=self.backbone, **kwargs)

    def download_imagenet(self):
        """ Downloads ImageNet weights and returns path to weights file.
        Weights can be downloaded at https://github.com/fizyr/keras-models/releases .
        """
        if self.backbone == 'vgg16':
            resource = keras.applications.vgg16.vgg16.WEIGHTS_PATH_NO_TOP
            checksum = '6d6bbae143d832006294945121d1f1fc'
        elif self.backbone == 'vgg19':
            resource = keras.applications.vgg19.vgg19.WEIGHTS_PATH_NO_TOP
            checksum = '253f8cb515780f3b799900260a226db6'
        else:
            raise ValueError("Backbone '{}' not recognized.".format(self.backbone))

        return get_file(
            '{}_weights_tf_dim_ordering_tf_kernels_notop.h5'.format(self.backbone),
            resource,
            cache_subdir='models',
            file_hash=checksum
        )

    def validate(self):
        """ Checks whether the backbone string is correct.
        """
        allowed_backbones = ['vgg16', 'vgg19']

        if self.backbone not in allowed_backbones:
            raise ValueError(
                'Backbone (\'{}\') not in allowed backbones ({}).'.format(self.backbone, allowed_backbones))

    def preprocess_image(self, inputs):
        """ Takes as input an image and prepares it for being passed through the network.
        """
        return preprocess_image(inputs, mode='caffe')


def vgg_retinanet(num_classes, backbone='vgg16', inputs=None, modifier=None, **kwargs):
    """ Constructs a retinanet model using a vgg backbone.

    Args
        num_classes: Number of classes to predict.
        backbone: Which backbone to use (one of ('vgg16', 'vgg19')).
        inputs: The inputs to the network (defaults to a Tensor of shape (None, None, 3)).
        modifier: A function handler which can modify the backbone before using it in retinanet (this can be used to freeze backbone layers for example).

    Returns
        RetinaNet model with a VGG backbone.
    """
    # choose default input
    if inputs is None:
        inputs = keras.layers.Input(shape=(None, None, 3))

    # create the vgg backbone
    if backbone == 'vgg16':
        vgg = keras.applications.VGG16(input_tensor=inputs, include_top=False, weights=None)
    elif backbone == 'vgg19':
        vgg = keras.applications.VGG19(input_tensor=inputs, include_top=False, weights=None)
    else:
        raise ValueError("Backbone '{}' not recognized.".format(backbone))

    if modifier:
        vgg = modifier(vgg)

    # create the full model
    layer_names = ["block3_pool", "block4_pool", "block5_pool"]
    layer_outputs = [vgg.get_layer(name).output for name in layer_names]
    return RetinaNetModel(inputs=inputs, num_classes=num_classes, backbone_layers=layer_outputs, **kwargs)



from notekeras.backend import backend as K
from notekeras.backend import keras

__all__ = ['AdamWarmup']


def identity(x):
    return x


symbolic = identity
if hasattr(K, 'symbolic'):
    symbolic = K.symbolic


class AdamWarmup(keras.optimizers.Optimizer):
    """Adam optimizer with warmup.

    Default parameters follow those provided in the original paper.

    # Arguments
        decay_steps: Learning rate will decay linearly to zero in decay steps.
        warmup_steps: Learning rate will increase linearly to lr in first warmup steps.
        learning_rate: float >= 0. Learning rate.
        beta_1: float, 0 < beta < 1. Generally close to 1.
        beta_2: float, 0 < beta < 1. Generally close to 1.
        epsilon: float >= 0. Fuzz factor. If `None`, defaults to `K.epsilon()`.
        weight_decay: float >= 0. Weight decay.
        weight_decay_pattern: A list of strings. The substring of weight names to be decayed.
                              All weights will be decayed if it is None.
        amsgrad: boolean. Whether to apply the AMSGrad variant of this
            algorithm from the paper "On the Convergence of Adam and
            Beyond".
    """

    def __init__(self,
                 decay_steps,
                 warmup_steps,
                 min_lr=0.0,
                 learning_rate=0.001,
                 beta_1=0.9,
                 beta_2=0.999,
                 epsilon=None,
                 weight_decay=0.,
                 weight_decay_pattern=None,
                 amsgrad=False,
                 name="",
                 **kwargs):
        learning_rate = kwargs.pop('lr', learning_rate)
        super(AdamWarmup, self).__init__(name=name, **kwargs)
        with K.name_scope(self.__class__.__name__):
            self.decay_steps = K.variable(decay_steps, name='decay_steps')
            self.warmup_steps = K.variable(warmup_steps, name='warmup_steps')
            self.min_lr = K.variable(min_lr, name='min_lr')
            self.iterations = K.variable(0, dtype='int64', name='iterations')
            self.learning_rate = K.variable(learning_rate, name='learning_rate')
            self.beta_1 = K.variable(beta_1, name='beta_1')
            self.beta_2 = K.variable(beta_2, name='beta_2')
            self.weight_decay = K.variable(weight_decay, name='weight_decay')
        if epsilon is None:
            epsilon = K.epsilon()
        self.epsilon = epsilon
        self.initial_weight_decay = weight_decay
        self.weight_decay_pattern = weight_decay_pattern
        self.amsgrad = amsgrad

    @property
    def lr(self):
        return self.learning_rate

    @lr.setter
    def lr(self, learning_rate):
        self.learning_rate = learning_rate

    @symbolic
    def get_updates(self, loss, params):
        grads = self.get_gradients(loss, params)
        self.updates = [K.update_add(self.iterations, 1)]

        t = K.cast(self.iterations, K.floatx()) + 1

        lr = K.switch(
            t <= self.warmup_steps,
            self.lr * (t / self.warmup_steps),
            self.min_lr + (self.lr - self.min_lr) * (1.0 - K.minimum(t, self.decay_steps) / self.decay_steps),
        )

        lr_t = lr * (K.sqrt(1. - K.pow(self.beta_2, t)) /
                     (1. - K.pow(self.beta_1, t)))

        ms = [K.zeros(K.int_shape(p), dtype=K.dtype(p), name='m_{}'.format(i)) for i, p in enumerate(params)]
        vs = [K.zeros(K.int_shape(p), dtype=K.dtype(p), name='v_{}'.format(i)) for i, p in enumerate(params)]
        if self.amsgrad:
            vhats = [K.zeros(K.int_shape(p), dtype=K.dtype(p), name='vh_{}'.format(i)) for i, p in enumerate(params)]
        else:
            vhats = [K.zeros(1, dtype=K.dtype(p), name='vh_{}'.format(i)) for i, p in enumerate(params)]
        self.weights = [self.iterations] + ms + vs + vhats

        for p, g, m, v, vhat in zip(params, grads, ms, vs, vhats):
            m_t = (self.beta_1 * m) + (1. - self.beta_1) * g
            v_t = (self.beta_2 * v) + (1. - self.beta_2) * K.square(g)
            if self.amsgrad:
                vhat_t = K.maximum(vhat, v_t)
                p_t = m_t / (K.sqrt(vhat_t) + self.epsilon)
                self.updates.append(K.update(vhat, vhat_t))
            else:
                p_t = m_t / (K.sqrt(v_t) + self.epsilon)

            if self.initial_weight_decay > 0.0:
                if self.weight_decay_pattern is None:
                    p_t += self.weight_decay * p
                else:
                    for pattern in self.weight_decay_pattern:
                        if pattern in p.name:
                            p_t += self.weight_decay * p
                            break
            p_t = p - lr_t * p_t

            self.updates.append(K.update(m, m_t))
            self.updates.append(K.update(v, v_t))
            new_p = p_t

            if getattr(p, 'constraint', None) is not None:
                new_p = p.constraint(new_p)

            self.updates.append(K.update(p, new_p))
        return self.updates

    def get_config(self):
        config = {
            'decay_steps': float(K.get_value(self.decay_steps)),
            'warmup_steps': float(K.get_value(self.warmup_steps)),
            'min_lr': float(K.get_value(self.min_lr)),
            'learning_rate': float(K.get_value(self.learning_rate)),
            'beta_1': float(K.get_value(self.beta_1)),
            'beta_2': float(K.get_value(self.beta_2)),
            'epsilon': self.epsilon,
            'weight_decay': float(K.get_value(self.weight_decay)),
            'weight_decay_pattern': self.weight_decay_pattern,
            'amsgrad': self.amsgrad,
        }
        base_config = super(AdamWarmup, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))



import tensorflow as tf
from tensorflow.keras.layers import ZeroPadding2D, Conv2D

from notekeras.component import Component
from notekeras.utils import compose


class BatchNormalization(tf.keras.layers.BatchNormalization):
    """
    "Frozen state" and "inference mode" are two separate concepts.
    `layer.trainable = False` is to freeze the layer, so the layer will use
    stored moving `var` and `mean` in the "inference mode", and both `gama`
    and `beta` will not be updated !
    """

    def call(self, x, training=False):
        if not training:
            training = tf.constant(False)
        training = tf.logical_and(training, self.trainable)
        return super().call(x, training)


def up_sample(input_layer):
    return tf.image.resize(input_layer, (input_layer.shape[1] * 2, input_layer.shape[2] * 2), method='bilinear')


def mish(x):
    return x * tf.math.tanh(tf.math.softplus(x))
    # return tf.keras.layers.Lambda(lambda x: x*tf.tanh(tf.math.log(1+tf.exp(x))))(x)


class YoloConv(Component):
    def __init__(self,
                 filters,
                 kernel_size,
                 down_sample=False,
                 activate=True,
                 bn=True,
                 activate_type='leaky',
                 *args, **kwargs):
        self.filters = filters
        self.kernel_size = kernel_size

        self.down_sample = down_sample
        self.activate = activate
        self.bn = bn
        self.activate_type = activate_type

        self.compose_list = None

        kwargs.update({"layer_depth": kwargs.get("layer_depth", 1)})
        super().__init__(*args, **kwargs)

    def build(self, input_shape):
        self.compose_list = []
        if self.down_sample:
            self.compose_list.append(ZeroPadding2D(((1, 0), (1, 0))))
            padding = 'valid'
            strides = 2
        else:
            strides = 1
            padding = 'same'

        self.compose_list.append(Conv2D(filters=self.filters,
                                        kernel_size=self.kernel_size,
                                        strides=strides,
                                        padding=padding,
                                        use_bias=not self.bn,
                                        kernel_regularizer=tf.keras.regularizers.l2(0.0005),
                                        kernel_initializer=tf.random_normal_initializer(stddev=0.01),
                                        bias_initializer=tf.constant_initializer(0.)))

        if self.bn:
            self.compose_list.append(BatchNormalization())

        self.layers.extend(self.compose_list)

    def call(self, inputs, **kwargs):
        output = compose(*self.compose_list)(inputs)

        if self.activate:
            if self.activate_type == "leaky":
                output = tf.nn.leaky_relu(output, alpha=0.1)
            elif self.activate_type == "mish":
                output = mish(output)

        return output

    def compute_output_shape(self, input_shape):
        return self.conv.compute_output_shape(input_shape)


class YoloResBlock(Component):
    def __init__(self, input_channel, filter_num1, filter_num2, name='YoloResidualBlock', activate_type='leaky', *args,
                 **kwargs):
        super(YoloResBlock, self).__init__(name=name)
        self.input_channel = input_channel
        self.filter_num1 = filter_num1
        self.filter_num2 = filter_num2
        self.activate_type = activate_type

        self.compose_list = []
        super(YoloResBlock, self).__init__(name=name, *args, **kwargs)

    def build(self, input_shape):
        self.compose_list = [YoloConv(filters=self.filter_num1, kernel_size=1, activate_type=self.activate_type),
                             YoloConv(filters=self.filter_num2, kernel_size=3, activate_type=self.activate_type)]

    def call(self, inputs, **kwargs):
        residual_output = inputs + compose(*self.compose_list)(inputs)
        return residual_output


def YoloCSPX(input_data, rec_num, filters=64, kernel_size=1, activate_type='mish'):
    route_data = YoloConv(filters=filters, kernel_size=kernel_size, activate_type=activate_type)(input_data)
    input_data = YoloConv(filters=filters, kernel_size=kernel_size, activate_type=activate_type)(input_data)

    for i in range(rec_num):
        input_data = YoloResBlock(input_channel=filters, filter_num1=filters, filter_num2=filters,
                                  activate_type=activate_type)(input_data)
    input_data = YoloConv(filters=filters, kernel_size=kernel_size, activate_type=activate_type)(input_data)

    return tf.concat([input_data, route_data], axis=-1)


def YoloNeck(conv, route_tmp, filters):
    conv = tf.concat([conv, route_tmp], axis=-1)

    conv = compose(YoloConv(filters=filters, kernel_size=1),
                   YoloConv(filters=filters * 2, kernel_size=3),
                   YoloConv(filters=filters, kernel_size=1),
                   YoloConv(filters=filters * 2, kernel_size=3),
                   YoloConv(filters=filters, kernel_size=1))(conv)

    return conv
